# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from brevitas.inject.enum import ScalingPerOutputType
from brevitas.proxy.groupwise_int_parameter_quant import GroupwiseWeightQuantProxyFromInjector
from brevitas.quant.base import *
from brevitas.quant.base import HQOActZeroPoint
from brevitas.quant.base import HQOWeightZeroPoint
from brevitas.quant.solver.act import ActQuantSolver
from brevitas.quant.solver.weight import WeightQuantSolver

__all__ = [
    'ShiftedUint8ActPerTensorFixedPoint',
    'ShiftedUint8ActPerTensorFloat',
    'ShiftedUint8WeightPerTensorFloat',
    'ShiftedUint8WeightPerChannelFloat',
    'ShiftedUint8ActPerTensorFixedPointMSE',
    'ShiftedUint8ActPerTensorFloatMSE',
    'ShiftedUint8WeightPerTensorFloatMSE',
    'ShiftedUint8WeightPerChannelFloatMSE',
    'ShiftedUint8ActPerTensorFloatHQO',
    'ShiftedUint8WeightPerChannelFloatHQO',
    'ShiftedUint8WeightPerTensorFloatHQO']


class ShiftedUint8ActPerTensorFixedPoint(ShiftedParamFromPercentileUintQuant,
                                         ParamFromRuntimePercentileIntervalScaling,
                                         PerTensorPoTScaling8bit,
                                         ActQuantSolver):
    """
    8-bit per-tensor unsigned int fixed-point activations quantizer with
    integer zero point. Both zero-point and scale factors are learned parameters initialized from
    runtime statistics.

        Examples:
        >>> from brevitas.nn import QuantReLU
        >>> act = QuantReLU(act_quant=ShiftedUint8ActPerTensorFixedPoint)
    """
    pass


class ShiftedUint8ActPerTensorFixedPointMSE(MSEAsymmetricScale,
                                            MSEActZeroPoint,
                                            ShiftedUint8ActPerTensorFixedPoint):
    """
    8-bit per-tensor unsigned int fixed-point activations quantizer with
    integer zero point. Both zero-point and scale factors are learned parameters initialized from
    MSE local loss.

        Examples:
        >>> from brevitas.nn import QuantReLU
        >>> act = QuantReLU(act_quant=ShiftedUint8ActPerTensorFixedPointMSE)
    """
    pass


class ShiftedUint8ActPerTensorFloat(ShiftedParamFromPercentileUintQuant,
                                    ParamFromRuntimePercentileIntervalScaling,
                                    PerTensorFloatScaling8bit,
                                    ActQuantSolver):
    """
    8-bit per-tensor unsigned int activations quantizer with floating-point scale factor and
    integer zero point. Both zero-point and scale factors are learned parameters initialized from
    runtime statistics.

        Examples:
        >>> from brevitas.nn import QuantReLU
        >>> act = QuantReLU(act_quant=ShiftedUint8ActPerTensorFloat)
    """
    pass


class ShiftedUint8ActPerTensorFloatMSE(MSEAsymmetricScale,
                                       MSEActZeroPoint,
                                       ShiftedUint8ActPerTensorFloat):
    """
    8-bit per-tensor unsigned int activations quantizer with floating-point scale factor and
    integer zero point. Both zero-point and scale factors are learned parameters initialized from
    MSE local loss.

        Examples:
        >>> from brevitas.nn import QuantReLU
        >>> act = QuantReLU(act_quant=ShiftedUint8ActPerTensorFloatMSE)
    """
    pass


class ShiftedUint8WeightPerTensorFloat(ShiftedMinUintQuant,
                                       MinMaxStatsScaling,
                                       PerTensorFloatScaling8bit,
                                       WeightQuantSolver):
    """
    8-bit per-tensor unsigned int weight quantizer with floating-point per-tensor scale factor and integer
    zero point. Both zero-point and scale factors are based on backpropagated statistics of the
    weight tensor.

    Examples:
        >>> from brevitas.nn import QuantLinear
        >>> fc = QuantLinear(10, 5, bias=False, weight_quant=ShiftedUint8WeightPerTensorFloat)
    """
    pass


class ShiftedUint8WeightPerTensorFloatMSE(MSEAsymmetricScale,
                                          MSEWeightZeroPoint,
                                          ShiftedUint8WeightPerTensorFloat):
    """
    8-bit per-tensor unsigned int weight quantizer with floating-point per-tensor scale factor and integer
    zero point. Both zero-point and scale factors are learned parameters initialized from MSE local losses.

    Examples:
        >>> from brevitas.nn import QuantLinear
        >>> fc = QuantLinear(10, 5, bias=False, weight_quant=ShiftedUint8WeightPerTensorFloatMSE)
    """
    pass


class ShiftedUint8WeightPerChannelFloat(ShiftedMinUintQuant,
                                        MinMaxStatsScaling,
                                        PerChannelFloatScaling8bit,
                                        WeightQuantSolver):
    """
    8-bit per-tensor unsigned int weight quantizer with floating-point per-channel scale factor and integer
    zero point. Both zero-point and scale factors are based on backpropagated statistics of the
    weight tensor.

    Examples:
        >>> from brevitas.nn import QuantLinear
        >>> fc = QuantLinear(10, 5, bias=False, weight_quant=ShiftedUint8WeightPerChannelFloat)
    """
    pass


class ShiftedUint8WeightPerChannelFloatMSE(MSEAsymmetricScale,
                                           MSEWeightZeroPoint,
                                           ShiftedUint8WeightPerChannelFloat):
    """
    8-bit per-tensor unsigned int weight quantizer with floating-point per-channel scale factor and integer
    zero point. Both zero-point and scale factors are learned parameters initialized from MSE local losses.

    Examples:
        >>> from brevitas.nn import QuantLinear
        >>> fc = QuantLinear(10, 5, bias=False, weight_quant=ShiftedUint8WeightPerChannelFloat)
    """
    pass


class ShiftedUint8WeightPerTensorFloatHQO(HQOWeightZeroPoint, ShiftedUint8WeightPerTensorFloat):
    """
    8-bit per-tensor unsigned int weight quantizer with floating-point per-channel scale factor and integer
    zero point. Zero-point is initialized from HQO local loss.

    Examples:
        >>> from brevitas.nn import QuantLinear
        >>> fc = QuantLinear(10, 5, bias=False, weight_quant=ShiftedUint8WeightPerTensorFloatHQO)
    """
    quantize_zero_point = False


class ShiftedUint8WeightPerChannelFloatHQO(HQOWeightZeroPoint, ShiftedUint8WeightPerChannelFloat):
    """
    8-bit per-tensor unsigned int weight quantizer with floating-point per-channel scale factor and integer
    zero point. Zero-point is initialized from HQO local loss.

    Examples:
        >>> from brevitas.nn import QuantLinear
        >>> fc = QuantLinear(10, 5, bias=False, weight_quant=ShiftedUint8WeightPerChannelFloatHQO)
    """
    quantize_zero_point = False


class ShiftedUint8WeightPerGroupFloatHQO(ShiftedUint8WeightPerChannelFloatHQO):
    """
    8-bit per-tensor unsigned int weight quantizer with floating-point per-channel scale factor and integer
    zero point.Zero-point is initialized from HQO local loss.
    Examples:
        >>> from brevitas.nn import QuantLinear
        >>> fc = QuantLinear(10, 5, bias=False, weight_quant=ShiftedUint8WeightPerChannelFloatHQO)
    """
    group_size = 32
    scaling_per_output_type = ScalingPerOutputType.GROUP
    proxy_class = GroupwiseWeightQuantProxyFromInjector


class ShiftedUint8ActPerTensorFloatHQO(HQOActZeroPoint, ShiftedUint8ActPerTensorFloat):
    """
    8-bit per-tensor unsigned int activations quantizer with floating-point scale factor and
    integer zero point. Zero-point is learned parameter initialized from
    HQO local loss.

        Examples:
        >>> from brevitas.nn import QuantReLU
        >>> act = QuantReLU(act_quant=ShiftedUint8ActPerTensorFloatHQO)
    """
    quantize_zero_point = False


class ShiftedUint8WeightGroupQuantFloat(ShiftedUint8WeightPerChannelFloat):
    """
    Block / group / vector signed asymmetric weight quantizer with float scales and zero-points.
    """
    proxy_class = GroupwiseWeightQuantProxyFromInjector
    scaling_per_output_type = ScalingPerOutputType.GROUP



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from dependencies import value

from brevitas.core.function_wrapper.ops_ste import CeilSte
from brevitas.core.scaling.runtime import RuntimeDynamicGroupStatsScaling
from brevitas.inject import ExtendedInjector
from brevitas.inject.enum import RestrictValueType
from brevitas.inject.enum import ScalingPerOutputType
from brevitas.proxy.groupwise_float_parameter_quant import \
    GroupwiseWeightFloatQuantProxyFromInjector
from brevitas.proxy.groupwise_float_runtime_quant import GroupwiseActFloatQuantProxyFromInjector
from brevitas.proxy.groupwise_int_parameter_quant import GroupwiseWeightQuantProxyFromInjector
from brevitas.proxy.groupwise_int_runtime_quant import GroupwiseActQuantProxyFromInjector
from brevitas.quant.base import IntQuant
from brevitas.quant.base import MaxStatsScaling
from brevitas.quant.base import MinMaxStatsScaling
from brevitas.quant.base import MSEAsymmetricScale
from brevitas.quant.base import MSESymmetricScale
from brevitas.quant.base import ShiftedMinUintQuant
from brevitas.quant.experimental.float_base import ScaledFloatActBase
from brevitas.quant.experimental.float_base import ScaledFloatWeightBase
from brevitas.quant.experimental.float_quant_ocp import FpOCPAct
from brevitas.quant.experimental.float_quant_ocp import FpOCPWeight
from brevitas.quant.solver.act import ActQuantSolver
from brevitas.quant.solver.weight import WeightQuantSolver


class GroupwiseWeightFloatProxyMixin(ExtendedInjector):
    proxy_class = GroupwiseWeightFloatQuantProxyFromInjector


class GroupwiseActFloatProxyMixin(ExtendedInjector):
    proxy_class = GroupwiseActFloatQuantProxyFromInjector


class GroupwiseWeightProxyMixin(ExtendedInjector):
    proxy_class = GroupwiseWeightQuantProxyFromInjector


class GroupwiseActProxyMixin(ExtendedInjector):
    proxy_class = GroupwiseActQuantProxyFromInjector


class MXWeightMixin(ExtendedInjector):
    group_size = 32
    restrict_scaling_type = RestrictValueType.POWER_OF_TWO
    restrict_value_float_to_int_impl = CeilSte
    scaling_per_output_type = ScalingPerOutputType.GROUP


class MXActMixin(ExtendedInjector):
    group_size = 32
    restrict_scaling_type = RestrictValueType.POWER_OF_TWO
    restrict_value_float_to_int_impl = CeilSte
    scaling_impl = RuntimeDynamicGroupStatsScaling
    scaling_per_output_type = ScalingPerOutputType.GROUP

    @value
    def stats_reduce_dim(group_dim):
        # If group_dim = -1, we need a workaround to avoid selecting wrong dim
        if group_dim == -1:
            return -1
        else:
            return group_dim + 1


class MXFloat8e4m3Weight(MXWeightMixin,
                         GroupwiseWeightFloatProxyMixin,
                         FpOCPWeight,
                         ScaledFloatWeightBase):
    """
    MX Float signed weight quantizer.
    """
    bit_width = 8
    exponent_bit_width = 4
    mantissa_bit_width = 3


class MXFloat8e4m3Act(MXActMixin, GroupwiseActFloatProxyMixin, FpOCPAct, ScaledFloatActBase):
    """
    MX Float signed activation quantizer.
    """
    bit_width = 8
    exponent_bit_width = 4
    mantissa_bit_width = 3


class MXFloat8e4m3WeightMSE(MXFloat8e4m3Weight, MSESymmetricScale):
    """
    MX Float signed weight quantizer with per-channel MSE-based scaling.
    """
    pass


class MXInt8Weight(MXWeightMixin,
                   GroupwiseWeightProxyMixin,
                   IntQuant,
                   MaxStatsScaling,
                   WeightQuantSolver):
    """
    MX Int signed weight quantizer.
    """
    bit_width = 8


class ShiftedMXUInt8Weight(MXWeightMixin,
                           GroupwiseWeightProxyMixin,
                           ShiftedMinUintQuant,
                           MinMaxStatsScaling,
                           WeightQuantSolver):
    """
    MX Int signed weight quantizer.
    """
    bit_width = 8


class MXInt8Act(MXActMixin, GroupwiseActProxyMixin, IntQuant, MaxStatsScaling, ActQuantSolver):
    """
    MX Int signed activation quantizer.
    """
    bit_width = 8


class MXInt8WeightMSE(MSESymmetricScale, MXInt8Weight):
    """
    MX Int signed weight quantizer with per-channel MSE-based scaling.
    """
    pass


class ShiftedMXUInt8WeightMSE(MSEAsymmetricScale, ShiftedMXUInt8Weight):
    """
    MX Int signed weight quantizer with per-channel MSE-based scaling.
    """
    pass



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

"""
ScriptModule wrappers for various variants of clamping.
"""
from typing import Optional, Tuple

import torch
from torch import Tensor
from torch.nn import Module

import brevitas
from brevitas.core.utils import StatelessBuffer
from brevitas.function import tensor_clamp
from brevitas.function.ops import max_float


class TensorClamp(brevitas.jit.ScriptModule):
    """
    ScriptModule wrapper for :func:`~brevitas.function.ops.tensor_clamp`.

    Examples:
        >>> tensor_clamp = TensorClamp()
        >>> min_val = torch.tensor(-2.0)
        >>> max_val = torch.tensor(2.0)
        >>> tensor_clamp(torch.tensor([-3.0, 3.0]), min_val, max_val)
        tensor([-2.,  2.])
    """

    def __init__(self) -> None:
        super(TensorClamp, self).__init__()

    @brevitas.jit.script_method
    def forward(self, x: Tensor, min_val: Tensor, max_val: Tensor):
        return tensor_clamp(x, min_val=min_val, max_val=max_val)


class ScalarClamp(brevitas.jit.ScriptModule):
    """
    ScriptModule wrapper for :func:`~torch.clamp`.

    Examples:
        >>> scalar_clamp = ScalarClamp(min_val=-2.0, max_val=2.0)
        >>> scalar_clamp(torch.tensor([-3.0, 3.0]))
        tensor([-2.,  2.])
    """

    __constants__ = ['min_val', 'max_val']

    def __init__(self, min_val, max_val) -> None:
        super(ScalarClamp, self).__init__()
        self.min_val = min_val
        self.max_val = max_val

    @brevitas.jit.script_method
    def forward(self, x: Tensor):
        return torch.clamp(x, min=self.min_val, max=self.max_val)


class ClampMin(brevitas.jit.ScriptModule):
    """
    ScriptModule wrapper for :func:`~torch.clamp_min`.

    Examples:
        >>> clamp_min = ClampMin(min_val=-2.0)
        >>> clamp_min(torch.tensor(-3.0))
        tensor(-2.)
    """

    __constants__ = ['min_val']

    def __init__(self, min_val: float) -> None:
        super(ClampMin, self).__init__()
        self.min_val = min_val

    @brevitas.jit.script_method
    def forward(self, x: Tensor):
        return x.clamp_min(self.min_val)


class FloatClamp(brevitas.jit.ScriptModule):
    """"
    ScriptModule for clamping minifloat formats to their inf/NaN implementations.

    Currently, inf/NaN codes have to be encoded through the mantissa.
    I.e. setting inf to 1101.111 (E4M3) is not a valid code.
    """

    __constants__ = ['saturating', 'inf_values', 'nan_values', 'signed']

    def __init__(
            self,
            tensor_clamp_impl: Module,
            signed: bool,
            inf_values: Optional[Tuple[str]] = None,
            nan_values: Optional[Tuple[str]] = None,
            max_available_float: Optional[Tensor] = None,
            saturating: bool = True,
            device: Optional[str] = None,
            dtype: Optional[torch.dtype] = None) -> None:
        super(FloatClamp, self).__init__()

        self.tensor_clamp_impl = tensor_clamp_impl
        self.saturating = saturating
        self.inf_values = inf_values
        self.nan_values = nan_values
        self.signed = signed

        if max_available_float:
            max_available_float = torch.tensor(max_available_float, device=device, dtype=dtype)
            self.max_available_float = StatelessBuffer(max_available_float)
        else:
            self.max_available_float = None

    def inf_nan_clamp(self, x, inf_mask, p_max_val_mask, n_max_val_mask):

        # if non-saturating, we need to map values greater than max_val to nan or inf
        if self.inf_values is not None:
            # we have inf values, so we set abs values > max_value to +- inf, and leave inf at inf
            x[p_max_val_mask] = torch.tensor(float('inf'))
            x[n_max_val_mask] = torch.tensor(float('-inf'))
        elif self.nan_values is not None:
            # no inf values, so we need to map them to NaN
            full_max_val_mask = torch.logical_or(p_max_val_mask, n_max_val_mask)
            x[full_max_val_mask] = torch.tensor(float('nan'))

            # we also map the inf values to NaN in this case
            x[inf_mask] = torch.tensor(float('nan'))
        else:
            raise RuntimeError(
                "Clamping is not saturating, but neither `inf_values` nor `nan_values` is specified"
            )
        return x

    def saturating_clamp(self, x, max_value, min_value):
        return self.tensor_clamp_impl(x, min_val=min_value, max_val=max_value)

    @brevitas.jit.script_method
    def forward(
            self,
            x: Tensor,
            exponent_bit_width: Tensor,
            mantissa_bit_width: Tensor,
            exponent_bias: Tensor):

        max_value = max_float(exponent_bit_width, mantissa_bit_width, exponent_bias)
        max_value = max_value if self.max_available_float is None else torch.min(
            max_value, self.max_available_float())
        min_value = torch.tensor(0.) if not self.signed else -max_value

        # Compute masks
        inf_mask = x.isinf()
        p_max_val_mask = x > max_value
        n_max_val_mask = -x > max_value

        # first clamp everything to +- max_value, basically the saturating case
        x = self.saturating_clamp(x, max_value, min_value)

        if not self.saturating:
            x = self.inf_nan_clamp(x, inf_mask, p_max_val_mask, n_max_val_mask)

        return x, self.saturating, self.inf_values, self.nan_values



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from packaging import version
import torch

from brevitas import torch_version

# This needs to be bumped until https://github.com/pytorch/pytorch/pull/94461 gets merged
if torch_version < version.parse('2.2'):
    import brevitas.backport.fx as fx
    from brevitas.backport.fx._compatibility import compatibility
    from brevitas.backport.fx._symbolic_trace import _assert_is_none
    from brevitas.backport.fx._symbolic_trace import _autowrap_check
    from brevitas.backport.fx._symbolic_trace import _find_proxy
    from brevitas.backport.fx._symbolic_trace import _is_fx_tracing_flag
    from brevitas.backport.fx._symbolic_trace import _orig_module_call
    from brevitas.backport.fx._symbolic_trace import _orig_module_getattr
    from brevitas.backport.fx._symbolic_trace import _patch_function
    from brevitas.backport.fx._symbolic_trace import _Patcher
    from brevitas.backport.fx._symbolic_trace import _proxyable_classes
    from brevitas.backport.fx._symbolic_trace import _wrapped_fns_to_patch
    from brevitas.backport.fx._symbolic_trace import _wrapped_methods_to_patch
    from brevitas.backport.fx._symbolic_trace import HAS_VARSTUFF
    from brevitas.backport.fx._symbolic_trace import map_aggregate
    from brevitas.backport.fx._symbolic_trace import PH
    from brevitas.backport.fx._symbolic_trace import Tracer
    from brevitas.backport.fx.graph import _PyTreeCodeGen
    from brevitas.backport.fx.graph import _PyTreeInfo
    from brevitas.backport.fx.graph import Graph
    from brevitas.backport.fx.graph import magic_methods
    from brevitas.backport.fx.graph import reflectable_magic_methods
    from brevitas.backport.fx.graph import Target
    from brevitas.backport.fx.graph_module import GraphModule
    from brevitas.backport.fx.immutable_collections import immutable_dict
    from brevitas.backport.fx.immutable_collections import immutable_list
    from brevitas.backport.fx.node import map_arg
    from brevitas.backport.fx.proxy import base_types
    from brevitas.backport.fx.proxy import Node
    from brevitas.backport.fx.proxy import ParameterProxy
    from brevitas.backport.fx.proxy import Proxy
    from brevitas.backport.fx.proxy import Scope
    from brevitas.backport.fx.proxy import ScopeContextManager
    import brevitas.backport.fx.traceback as fx_traceback
else:
    from torch.fx import Graph
    from torch.fx import GraphModule
    from torch.fx import map_arg
    from torch.fx import Node
    from torch.fx import Proxy
    from torch.fx import Tracer
    from torch.fx.graph import magic_methods
    from torch.fx.graph import reflectable_magic_methods
    from torch.fx.graph import Target
    from torch.fx.proxy import base_types
    from torch.fx._symbolic_trace import _orig_module_call
    from torch.fx._symbolic_trace import _orig_module_getattr
    from torch.fx._symbolic_trace import _autowrap_check
    from torch.fx._symbolic_trace import _Patcher, map_aggregate
    from torch.fx._symbolic_trace import _wrapped_fns_to_patch, _wrapped_methods_to_patch
    from torch.fx._symbolic_trace import _find_proxy, _patch_function, HAS_VARSTUFF
    from torch.fx.immutable_collections import immutable_dict, immutable_list
    import torch.fx as fx
    from torch.fx._compatibility import compatibility
    from torch.fx._symbolic_trace import _assert_is_none
    from torch.fx._symbolic_trace import PH
    from torch.fx.graph import _PyTreeCodeGen
    from torch.fx.graph import _PyTreeInfo
    from torch.fx.proxy import ParameterProxy
    from torch.fx.proxy import Scope
    from torch.fx.proxy import ScopeContextManager
    import torch.fx.traceback as fx_traceback
    from torch.fx._symbolic_trace import _is_fx_tracing_flag
    from torch.fx._symbolic_trace import _proxyable_classes

from .brevitas_tracer import brevitas_symbolic_trace
from .brevitas_tracer import brevitas_value_trace
from .brevitas_tracer import symbolic_trace
from .brevitas_tracer import value_trace



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import torch
from torch import Tensor
import torch.nn as nn

from brevitas.core.scaling import AccumulatorAwareParameterPreScaling
from brevitas.core.scaling import AccumulatorAwareZeroCenterParameterPreScaling
import brevitas.nn as qnn
from brevitas.nn.quant_layer import QuantWeightBiasInputOutputLayer as QuantWBIOL

EPS = 1e-10


def _get_a2q_module(module: nn.Module):
    for submod in module.modules():
        if isinstance(submod, AccumulatorAwareParameterPreScaling):
            return submod
    return None


def _calc_a2q_acc_bit_width(
        weight_max_l1_norm: Tensor, input_bit_width: Tensor, input_is_signed: bool):
    """Using the closed-form bounds on accumulator bit-width as derived in
    `A2Q: Accumulator-Aware Quantization with Guaranteed Overflow Avoidance`.
    This function returns the minimum accumulator bit-width that can be used
    without risk of overflow."""
    assert weight_max_l1_norm.numel() == 1
    input_is_signed = float(input_is_signed)
    weight_max_l1_norm = torch.clamp_min(weight_max_l1_norm, EPS)
    alpha = torch.log2(weight_max_l1_norm) + input_bit_width - input_is_signed
    phi = lambda x: torch.log2(1. + pow(2., -x))
    min_bit_width = alpha + phi(alpha) + 1.
    min_bit_width = torch.ceil(min_bit_width)
    return min_bit_width


def _calc_a2q_plus_acc_bit_width(
        weight_max_l1_norm: Tensor, input_bit_width: Tensor, input_is_signed: bool):
    """Using the closed-form bounds on accumulator bit-width as derived in `A2Q+:
    Improving Accumulator-Aware Weight Quantization`. This function returns the
    minimum accumulator bit-width that can be used without risk of overflow,
    assuming that the floating-point weights are zero-centered."""
    input_is_signed = float(input_is_signed)
    assert weight_max_l1_norm.numel() == 1
    weight_max_l1_norm = torch.clamp_min(weight_max_l1_norm, EPS)
    input_range = pow(2., input_bit_width) - 1.  # 2^N - 1.
    min_bit_width = torch.log2(weight_max_l1_norm * input_range + 2.)
    min_bit_width = torch.ceil(min_bit_width)
    return min_bit_width


def _calc_min_acc_bit_width(module: QuantWBIOL) -> Tensor:
    assert isinstance(module, qnn.QuantConv2d), "Error: function only support QuantConv2d."

    # bit-width and sign need to come from the quant tensor of the preceding layer if no io_quant
    input_bit_width = module.input_quant.bit_width()
    input_is_signed = float(module.input_quant.is_signed)

    # the tensor quantizer requires a QuantTensor with specified bit-width and sign
    quant_weight = module.quant_weight()
    quant_weight = quant_weight.int().float()
    quant_weight_per_channel_l1_norm = quant_weight.norm(p=1, dim=(1, 2, 3))

    # using the closed-form bounds on accumulator bit-width
    min_bit_width = _calc_a2q_acc_bit_width(
        quant_weight_per_channel_l1_norm.max(),
        input_bit_width=input_bit_width,
        input_is_signed=input_is_signed)
    if isinstance(_get_a2q_module(module), AccumulatorAwareZeroCenterParameterPreScaling):
        min_bit_width = _calc_a2q_plus_acc_bit_width(
            quant_weight_per_channel_l1_norm.max(),
            input_bit_width=input_bit_width,
            input_is_signed=input_is_signed)
    return min_bit_width


def evaluate_accumulator_bit_widths(model: nn.Module, inp: Tensor):
    model(inp)  # collect quant inputs now that caching is enabled
    stats = dict()
    for name, module in model.named_modules():
        # ESPCN only has quantized conv2d nodes and the last one (i.e., conv4.conv) is decoupled
        # from the input quantizer. Will check for more layer types in the future with other
        # example models with different neural architectures.
        if isinstance(module, qnn.QuantConv2d):
            acc_bit_width = _calc_min_acc_bit_width(module)
            stats[name] = acc_bit_width.item()
    return stats



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Optional, Tuple

import torch
import torch.nn as nn

import brevitas
from brevitas.core.function_wrapper import RoundSte
from brevitas.core.scaling import ConstScaling
from brevitas.core.utils import StatelessBuffer
from brevitas.utils.torch_utils import float_internal_scale


class FloatQuant(brevitas.jit.ScriptModule):
    __constants__ = ['signed', 'eps']

    def __init__(
            self,
            bit_width: int,
            signed: bool,
            exponent_bit_width: int,
            mantissa_bit_width: int,
            exponent_bias: int,
            float_clamp_impl: nn.Module,
            input_view_impl: nn.Module,
            scaling_impl: Optional[nn.Module] = None,
            float_scaling_impl: Optional[nn.Module] = None,
            float_to_int_impl: nn.Module = RoundSte(),
            device: Optional[str] = None,
            dtype: Optional[torch.dtype] = None):
        super(FloatQuant, self).__init__()
        if bit_width != exponent_bit_width + mantissa_bit_width + int(signed):
            raise RuntimeError("Mismatch between total bit-width, exponent, mantissa and sign.")
        self.bit_width = StatelessBuffer(torch.tensor(float(bit_width), device=device, dtype=dtype))
        self.signed: bool = signed
        self.float_to_int_impl = float_to_int_impl
        if exponent_bit_width == 0:
            raise RuntimeError("Exponent bit width cannot be 0.")
        self.exponent_bit_width = StatelessBuffer(
            torch.tensor(float(exponent_bit_width), device=device, dtype=dtype))
        if mantissa_bit_width == 0:
            raise RuntimeError("Mantissa bit width cannot be 0.")
        self.mantissa_bit_width = StatelessBuffer(
            (torch.tensor(float(mantissa_bit_width), device=device, dtype=dtype)))
        self.exponent_bias = StatelessBuffer(
            torch.tensor(float(exponent_bias), device=device, dtype=dtype))

        self.fp_internal_scale_min = StatelessBuffer(
            1. - self.exponent_bias() - self.mantissa_bit_width())

        if scaling_impl is None:
            scaling_impl = ConstScaling(1., device=device, dtype=dtype)

        self.input_view_impl = input_view_impl
        # Zero-point is currently hardcoded to 0
        self.zero_point_impl = StatelessBuffer(torch.tensor(0., device=device, dtype=dtype))
        self.float_scaling_impl = float_scaling_impl
        self.scaling_impl = scaling_impl
        self.float_clamp_impl = float_clamp_impl

        # To avoid log(0), we add small a small value based on the used dtype
        if dtype is None:
            dtype = torch.get_default_dtype()
        self.eps = torch.finfo(dtype).tiny
        self.observer_only = brevitas.jit.Attribute(False, bool)

    @brevitas.jit.script_method
    def quantize(self, x: torch.Tensor, scale: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        x = self.input_view_impl(x)
        scaled_x = x / scale
        internal_scale = float_internal_scale(
            scaled_x, self.mantissa_bit_width(), self.fp_internal_scale_min(), self.eps)
        val_fp_quant = internal_scale * self.float_to_int_impl(scaled_x / internal_scale)
        return val_fp_quant, scale

    @brevitas.jit.script_method
    def dequantize(self, y, scale):
        return y * scale

    @brevitas.jit.script_method
    def forward(self, x):
        if self.float_scaling_impl is not None:
            float_scaling_impl_value = self.float_scaling_impl(
                self.exponent_bit_width(), self.mantissa_bit_width(), self.exponent_bias())
        else:
            float_scaling_impl_value = None
        scale = self.scaling_impl(x, float_scaling_impl_value)
        if self.observer_only:
            y = x
            saturating, inf_values, nan_values = self.float_clamp_impl.saturating, self.float_clamp_impl.inf_values, self.float_clamp_impl.nan_values
        else:
            y, scale = self.quantize(x, scale)
            # after quantizing, clamp to special cases like NaN/inf if they are set
            y, saturating, inf_values, nan_values = self.float_clamp_impl(
                y, self.exponent_bit_width(), self.mantissa_bit_width(), self.exponent_bias())
            y = self.dequantize(y, scale)
        # This is to respect the current interface of proxies
        return y, scale, self.zero_point_impl(), self.exponent_bit_width(), self.mantissa_bit_width(), self.exponent_bias(), saturating, inf_values, nan_values



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABCMeta
from typing import Optional

from torch import nn
from torch import tensor
from typing_extensions import Protocol
from typing_extensions import runtime_checkable

from brevitas import config
from brevitas.common import ExportMixin
from brevitas.core.scaling import ScalingPerOutputType
from brevitas.core.utils import StatelessBuffer
from brevitas.inject import BaseInjector as Injector
from brevitas.utils.quant_utils import float_to_int_impl_to_enum

__all__ = [
    'QuantProxyProtocol',
    'QuantProxyFromInjector',]


def _is_groupwise(quant_injector):
    return 'scaling_per_output' in quant_injector and quant_injector.scaling_per_output == ScalingPerOutputType.GROUP


def _is_narrow_range(quant_injector):
    if 'narrow_range' in quant_injector:
        return quant_injector.narrow_range
    return None


def _rounding_mode(quant_injector):
    if 'float_to_int_impl_type' in quant_injector:
        return str(quant_injector.float_to_int_impl_type)
    elif 'float_to_int_impl' in quant_injector:
        try:
            impl_type = float_to_int_impl_to_enum(quant_injector.float_to_int_impl)
            return str(impl_type).upper()
        except:
            return None
    else:
        return None


def _update_state_dict_impl(quant_injector):
    try:
        impl = quant_injector.update_state_dict_impl
    except:
        impl = None
    return impl


@runtime_checkable
class QuantProxyProtocol(Protocol):
    is_quant_enabled: bool
    is_signed: Optional[bool]
    is_narrow_range: Optional[bool]
    rounding_mode: Optional[str]

    def add_tracked_module(self, module: nn.Module) -> None:
        ...


class QuantProxyFromInjector(ExportMixin, nn.Module, QuantProxyProtocol):
    __metaclass__ = ABCMeta

    def __init__(self, quant_layer: nn.Module, quant_injector: Injector) -> None:
        ExportMixin.__init__(self)
        nn.Module.__init__(self)
        QuantProxyProtocol.__init__(self)
        self.update_state_dict_impl = _update_state_dict_impl(quant_injector)
        self.quant_injector = quant_injector
        self.quant_injector = quant_injector.let(proxy_module=self)
        self._zero_hw_sentinel = StatelessBuffer(tensor(0.0))
        self.tensor_quant = None
        # Use a normal list and not a ModuleList since this is a pointer to parent modules
        self.tracked_module_list = []
        self.add_tracked_module(quant_layer)
        self.disable_quant = False
        # Torch.compile compatibility requires this
        self.is_signed = quant_injector.signed if 'signed' in quant_injector else None

    @property
    def requires_export_handler(self):
        return self.is_quant_enabled

    def update_tracked_modules(self):
        """Update the modules tracked by the injector with the modules tracked by the proxy"""
        if len(self.tracked_module_list) == 1:
            self.quant_injector = self.quant_injector.let(module=self.tracked_module_list[0])
        else:
            # set the list in the injector as a tuple to avoid dealing with inplace modifications
            self.quant_injector = self.quant_injector.let(module=tuple(self.tracked_module_list))

    def init_tensor_quant(self):
        self.tensor_quant = self.quant_injector.tensor_quant

    @property
    def is_quant_enabled(self):
        return not self.disable_quant and self.tensor_quant is not None

    @property
    def is_groupwise(self):
        return _is_groupwise(self.quant_injector)

    @property
    def is_narrow_range(self):
        return _is_narrow_range(self.quant_injector)

    @property
    def rounding_mode(self):
        return _rounding_mode(self.quant_injector)

    def add_tracked_module(self, module: nn.Module) -> None:
        if module is not None:
            self.tracked_module_list.append(module)
            self.update_tracked_modules()
            self.init_tensor_quant()
        else:
            raise RuntimeError("Trying to add None as a parent module.")

    def apply_input_view(self, x):
        return self.input_view_impl(x)

    def _load_from_state_dict(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys,
            error_msgs):
        if self.update_state_dict_impl is not None:
            self.update_state_dict_impl(prefix, state_dict)
        super(QuantProxyFromInjector, self)._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)
        # reload tensor_quant on changes of the state_dict
        # this is called after the parent module state_dict is restored (e.g. weights)
        # so init_tensor_quant takes into account new data from the parent module,
        # but before the state_dict of tensor_quant is loaded, so in case e.g. there is a value
        # for the parameter already, it's not overwritten
        if config.REINIT_ON_STATE_DICT_LOAD:
            self.init_tensor_quant()
        # for retrocompatibility with when it wasn't removed
        zero_hw_sentinel_key = prefix + 'zero_hw_sentinel'
        if zero_hw_sentinel_key in unexpected_keys:
            unexpected_keys.remove(zero_hw_sentinel_key)



# Copyright (C) 2024, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from brevitas.quant import Int8AccumulatorAwareWeightQuant
from brevitas.quant import Int8AccumulatorAwareZeroCenterWeightQuant

__all__ = ["CommonIntAccumulatorAwareWeightQuant", "CommonIntAccumulatorAwareZeroCenterWeightQuant"]

SCALING_MIN_VAL = 1e-8


class CommonIntAccumulatorAwareWeightQuant(Int8AccumulatorAwareWeightQuant):
    """A2Q: Accumulator-Aware Quantization with Guaranteed Overflow Avoidance"""
    bit_width = None
    scaling_min_val = SCALING_MIN_VAL
    pre_scaling_min_val = SCALING_MIN_VAL


class CommonIntAccumulatorAwareZeroCenterWeightQuant(Int8AccumulatorAwareZeroCenterWeightQuant):
    """A2Q+: Improving Accumulator-Aware Weight Quantization"""
    bit_width = None
    scaling_min_val = SCALING_MIN_VAL
    pre_scaling_min_val = SCALING_MIN_VAL



from typing import Optional, Tuple

import torch
from torch import nn

from brevitas.nn.equalized_layer import EqualizedModule
from brevitas.utils.torch_utils import KwargsForwardHook


def attention_mask_handler(
        attention_mask, batch_size, num_heads, query_seq_length, key_value_seq_length):
    """Re-arrange attention mask to go from 4D to 3D (explicit batch_size and n_heads) or 2D
    (implicit batch_size and n_heads)."""
    if len(attention_mask.shape) == 4:
        if attention_mask.shape[0] == 1:
            attention_mask = attention_mask.repeat(batch_size, 1, 1, 1)
        if attention_mask.shape[1] == 1:
            attention_mask = attention_mask.repeat(1, num_heads, 1, 1)
        if attention_mask.shape[2] == 1:
            attention_mask = attention_mask.repeat(1, 1, query_seq_length, 1)
        attention_mask = attention_mask.view(
            batch_size * num_heads, query_seq_length, key_value_seq_length)
    elif len(attention_mask.shape) == 2 and attention_mask.shape[0] == 1:
        # This could happen in Encoder-like architecture
        assert query_seq_length == key_value_seq_length
        attention_mask = attention_mask.repeat(query_seq_length, 1)
    return attention_mask


class MultiheadAttentionWrapper(nn.Module):

    def __init__(
            self,
            embed_dim,
            num_heads,
            dropout=0.,
            bias=True,
            add_bias_kv=False,
            add_zero_attn=False,
            kdim=None,
            vdim=None,
            batch_first=False,
            device=None,
            dtype=None) -> None:
        super().__init__()
        self.mha = nn.MultiheadAttention(
            embed_dim,
            num_heads,
            dropout,
            bias,
            add_bias_kv,
            add_zero_attn,
            kdim,
            vdim,
            batch_first,
            device,
            dtype)

    @property
    def wrapped_mha(self):
        mha = self.mha
        # Workaround for activation equalization for when mha is wrapped
        # KwargsForwardHook is inserted during act equalization
        # EqualizedModule is inserted after act equalization
        if isinstance(mha, KwargsForwardHook):
            mha = mha.module
        if isinstance(mha, EqualizedModule):
            mha = mha.layer
        return mha

    @property
    def num_heads(self):
        return self.wrapped_mha.num_heads

    @property
    def batch_first(self):
        return self.wrapped_mha.batch_first

    def _load_from_state_dict(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys,
            error_msgs):

        def set_bias(value):
            bias_name = f'{prefix}mha.in_proj_bias'
            if bias_name in state_dict:
                state_dict[bias_name] += value
            else:
                state_dict[bias_name] = value

        def set_weight(value):
            weight_name = f'{prefix}mha.in_proj_weight'
            if weight_name in state_dict:
                state_dict[weight_name] += value
            else:
                state_dict[weight_name] = value

        embed_dim = self.mha.embed_dim
        for name, value in list(state_dict.items()):
            if prefix + 'q_proj.weight' in name:
                weight = torch.zeros((3 * embed_dim, embed_dim),
                                     device=value.device,
                                     dtype=value.dtype)
                weight[:embed_dim] = value
                set_weight(weight)
                del state_dict[name]
            elif prefix + 'k_proj.weight' in name:
                weight = torch.zeros((3 * embed_dim, embed_dim),
                                     device=value.device,
                                     dtype=value.dtype)
                weight[embed_dim:2 * embed_dim] = value
                set_weight(weight)
                del state_dict[name]
            elif prefix + 'v_proj.weight' in name:
                weight = torch.zeros((3 * embed_dim, embed_dim),
                                     device=value.device,
                                     dtype=value.dtype)
                weight[2 * embed_dim:3 * embed_dim] = value
                set_weight(weight)
                del state_dict[name]
            if prefix + 'q_proj.bias' in name:
                bias = torch.zeros(3 * embed_dim, device=value.device, dtype=value.dtype)
                bias[:embed_dim] = value
                set_bias(bias)
                del state_dict[name]
            elif prefix + 'k_proj.bias' in name:
                bias = torch.zeros(3 * embed_dim, device=value.device, dtype=value.dtype)
                bias[embed_dim:2 * embed_dim] = value
                set_bias(bias)
                del state_dict[name]
            elif prefix + 'v_proj.bias' in name:
                bias = torch.zeros(3 * embed_dim, device=value.device, dtype=value.dtype)
                bias[2 * embed_dim:3 * embed_dim] = value
                set_bias(bias)
                del state_dict[name]
            elif prefix + 'out_proj.weight' in name:
                state_dict[prefix + 'mha.out_proj.weight'] = value
                del state_dict[name]
            elif prefix + 'out_proj.bias' in name:
                state_dict[prefix + 'mha.out_proj.bias'] = value
                del state_dict[name]
        return super()._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)


class QuantizableOPTAttention(MultiheadAttentionWrapper):

    def forward(
        self,
        hidden_states: torch.Tensor,
        key_value_states: Optional[torch.Tensor] = None,
        past_key_value: Optional[Tuple[torch.Tensor]] = None,
        attention_mask: Optional[torch.Tensor] = None,
        layer_head_mask: Optional[torch.Tensor] = None,
        output_attentions: bool = False,
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[Tuple[torch.Tensor]]]:
        if key_value_states is None:
            key_value_states = hidden_states
        if layer_head_mask is not None:
            raise RuntimeError("layer_head_mask is not supported.")
        if self.batch_first:
            batch_size, query_seq_length = hidden_states.shape[:2]
            key_value_seq_length = key_value_states.shape[1]
        else:
            query_seq_length, batch_size = hidden_states.shape[:2]
            key_value_seq_length = key_value_states.shape[0]
        num_heads = self.num_heads
        attention_mask = attention_mask_handler(
            attention_mask, batch_size, num_heads, query_seq_length, key_value_seq_length)
        attn_output, attn_output_weights = self.mha(
            hidden_states,
            key_value_states,
            key_value_states,
            attn_mask=attention_mask,
            need_weights=output_attentions,
            average_attn_weights=False)
        past_key_value = None
        return attn_output, attn_output_weights, past_key_value



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Optional

from torch import Tensor
import torch.nn as nn

from brevitas.core.restrict_val import FloatRestrictValue
from brevitas.core.restrict_val import RestrictValueType
from brevitas.core.scaling import ScalingImplType
import brevitas.nn as qnn
from brevitas.nn.quant_layer import WeightQuantType
from brevitas.quant import Int8AccumulatorAwareWeightQuant
from brevitas.quant import Int8AccumulatorAwareZeroCenterWeightQuant
from brevitas.quant import Int8ActPerTensorFloat
from brevitas.quant import Int8WeightPerTensorFloat
from brevitas.quant import Uint8ActPerTensorFloat


class CommonIntWeightPerChannelQuant(Int8WeightPerTensorFloat):
    """
    Common per-channel weight quantizer with bit-width set to None so that it's forced to be
    specified by each layer.
    """
    scaling_per_output_channel = True


class CommonIntAccumulatorAwareWeightQuant(Int8AccumulatorAwareWeightQuant):
    """A2Q: Accumulator-Aware Quantization with Guaranteed Overflow Avoidance"""
    restrict_scaling_impl = FloatRestrictValue  # backwards compatibility
    bit_width = None


class CommonIntAccumulatorAwareZeroCenterWeightQuant(Int8AccumulatorAwareZeroCenterWeightQuant):
    """A2Q+: Improving Accumulator-Aware Weight Quantization"""
    bit_width = None


class CommonIntActQuant(Int8ActPerTensorFloat):
    """
    Common signed act quantizer with bit-width set to None so that it's forced to be specified by
    each layer.
    """
    bit_width = None
    restrict_scaling_type = RestrictValueType.LOG_FP


class CommonUintActQuant(Uint8ActPerTensorFloat):
    """Common unsigned act quantizer with bit-width set to None so that it's forced to be
    specified by each layer"""
    bit_width = None
    restrict_scaling_type = RestrictValueType.LOG_FP


class ConstUint8ActQuant(CommonUintActQuant):
    """8-bit unsigned integer activation quantizer with constant unit scaling factor, used
    by the models to quantize outputs into the image space"""
    scaling_impl_type = ScalingImplType.CONST
    scaling_init = 1.


class QuantNearestNeighborConvolution(nn.Module):
    """Quantized nearest neighbor resize convolution"""

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: Optional[int] = 3,
            stride: Optional[int] = 1,
            padding: Optional[int] = 1,
            upscale_factor: Optional[int] = 3,
            signed_act: Optional[bool] = False,
            bias: Optional[bool] = True,
            weight_quant: WeightQuantType = CommonIntWeightPerChannelQuant,
            acc_bit_width: Optional[int] = 32,
            act_bit_width: Optional[int] = 8,
            weight_bit_width: Optional[int] = 8):
        super().__init__()

        # Using unsigned int activation quantization if the preceding layer has
        # a non-negative range (e.g., following a ReLU activation function)
        act_quant = CommonIntActQuant if signed_act else CommonUintActQuant

        self.upscale_factor = upscale_factor
        # Need to have the quantization node before the nearest neighbor upsampling node
        # for FINN compatibility since the FINN compiler will streamline the quantization
        # node with the preceding monotonic activation function. In the case of ESPCN, this
        # is a ReLU. We need to return the QuantTensor though so that the conv2d is aware
        # of the input bit-width for accumulator-aware quantization (A2Q). For more discussion
        # on this, see https://arxiv.org/abs/2301.13376.
        self.input_quant = qnn.QuantIdentity(
            act_quant=act_quant, return_quant_tensor=True, bit_width=act_bit_width)
        self.interp = qnn.QuantUpsamplingNearest2d(
            scale_factor=upscale_factor, return_quant_tensor=True)
        self.conv = qnn.QuantConv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            padding=padding,
            stride=stride,
            bias=bias,
            input_quant=None,
            weight_accumulator_bit_width=acc_bit_width,
            weight_bit_width=weight_bit_width,
            weight_quant=weight_quant)

    def forward(self, inp: Tensor) -> Tensor:
        return self.conv(self.interp(self.input_quant(inp)))



# Copyright 2024 Advanced Micro Devices, Inc
#
# Licensed under the Apache License v2.0 with LLVM Exceptions.
# See https://llvm.org/LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from pathlib import Path

import torch

from shark_turbine import aot

from ..model import Unet2DConditionModel, ClassifierFreeGuidanceUnetModel




def main():
    from ....utils import cli

    parser = cli.create_parser()
    cli.add_input_dataset_options(parser)
    parser.add_argument("--device", default="cuda:0", help="Torch device to run on")
    parser.add_argument("--dtype", default="float16", help="DType to run in")
    parser.add_argument("--export", type=Path, help="Export to path (vs run)")
    parser.add_argument(
        "--inputs",
        type=Path,
        help="Safetensors file of inputs (or random if not given)",
    )
    args = cli.parse(parser)

    device = args.device
    dtype = getattr(torch, args.dtype)

    ds = cli.get_input_dataset(args)
    ds.to(device=device)

    cond_unet = Unet2DConditionModel.from_dataset(ds)
    mdl = ClassifierFreeGuidanceUnetModel(cond_unet)

    # Run a step for debugging.
    if args.inputs:
        inputs = load_inputs(args.inputs, dtype=dtype, device=device)
    else:
        inputs = get_random_inputs(dtype=dtype, device=device)

    if args.export:
        # Temporary: Need a dedicated exporter.
        output = aot.export(
            mdl,
            kwargs=inputs,
        )
        output.save_mlir(args.export)
    else:
        results = mdl.forward(**inputs)
        print("1-step results:", results)


if __name__ == "__main__":
    main()


# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC
from abc import ABCMeta
from abc import abstractmethod
from typing import Any, List, Optional, Tuple, Union
from warnings import warn

import torch
from torch import Tensor
import torch.nn as nn
from typing_extensions import Protocol
from typing_extensions import runtime_checkable

from brevitas import config
from brevitas import is_dynamo_compiling
from brevitas.core.function_wrapper.misc import Identity
from brevitas.function import max_int
from brevitas.inject import BaseInjector as Injector
from brevitas.quant_tensor import _unpack_quant_tensor
from brevitas.quant_tensor import IntQuantTensor
from brevitas.quant_tensor import QuantTensor
from brevitas.utils.quant_utils import _CachedIO
from brevitas.utils.torch_utils import compute_channel_view_shape

from .quant_proxy import QuantProxyFromInjector
from .quant_proxy import QuantProxyProtocol

__all__ = [
    'WeightQuantProxyFromInjector',
    'WeightQuantProxyFromInjectorBase',
    'BiasQuantProxyFromInjector',
    'BiasQuantProxyFromInjectorBase',
    'WeightQuantProxyProtocol',
    'BiasQuantProxyProtocol']


@runtime_checkable
class WeightQuantProxyProtocol(QuantProxyProtocol, Protocol):

    def forward(self, x: torch.Tensor) -> QuantTensor:
        ...


@runtime_checkable
class BiasQuantProxyProtocol(QuantProxyProtocol, Protocol):
    requires_input_scale: bool

    def forward(
            self, x: Tensor, input_scale: Optional[Tensor],
            input_bit_width: Optional[Tensor]) -> QuantTensor:
        ...


class ParameterQuantProxyFromInjector(QuantProxyFromInjector):
    __metaclass__ = ABCMeta

    @property
    @abstractmethod
    def tracked_parameter_list(self):
        pass

    def init_tensor_quant(self, preserve_state_dict=False):
        param_list = self.tracked_parameter_list

        # params might not be there yet, e.g. bias before merging
        if param_list:
            if preserve_state_dict:
                reinit_on_state_dict = config.REINIT_ON_STATE_DICT_LOAD
                ignore_missing_key = config.IGNORE_MISSING_KEYS
                config.REINIT_ON_STATE_DICT_LOAD = False
                config.IGNORE_MISSING_KEYS = True
                state_dict = self.state_dict()
            self.quant_injector = self.quant_injector.let(tracked_parameter_list=param_list)
            super(ParameterQuantProxyFromInjector, self).init_tensor_quant()
            if preserve_state_dict:
                self.load_state_dict(state_dict)
                config.IGNORE_MISSING_KEYS = ignore_missing_key
                config.REINIT_ON_STATE_DICT_LOAD = reinit_on_state_dict

    def max_uint_value(self, bit_width):
        return max_int(False, self.is_narrow_range, bit_width)


class WeightQuantProxyFromInjectorBase(ParameterQuantProxyFromInjector,
                                       WeightQuantProxyProtocol,
                                       ABC):

    def __init__(self, quant_layer: nn.Module, quant_injector: Injector) -> None:
        super().__init__(quant_layer, quant_injector)
        self._cached_weight = None
        self._cache_inference_quant_weight = False
        self.cache_inference_quant_weight_metadata_only = False
        self.cache_class = None  # To be redefined by each class
        self.quant_tensor_class = None  # To be redefined by each class

    @property
    def input_view_impl(self):
        if self.tensor_quant is not None:
            return self.tensor_quant.int_quant.input_view_impl
        else:
            return Identity()

    @property
    def cache_inference_quant_weight(self):
        return self._cache_inference_quant_weight

    @cache_inference_quant_weight.setter
    def cache_inference_quant_weight(self, value):
        if not value:
            self._cached_weight = None
        self._cache_inference_quant_weight = value

    @property
    def tracked_parameter_list(self):
        return [m.weight for m in self.tracked_module_list if m.weight is not None]

    @property
    def requires_quant_input(self):
        return False

    @abstractmethod
    def create_quant_tensor(self, qt_args: Tuple[Any]) -> Union[Tensor, QuantTensor]:
        raise NotImplementedError

    def forward(self, x: torch.Tensor) -> Union[Tensor, QuantTensor]:
        if self.is_quant_enabled:
            # If quant is enabled the priority is:
            # - export mode
            # - quantization flow
            if self.export_mode:
                out = self.export_handler(x)
                if is_dynamo_compiling():
                    out = out[0]
                else:
                    out = self.create_quant_tensor(out)
            else:
                out = self.tensor_quant(x)
                if is_dynamo_compiling():
                    out = out[0]
                else:
                    out = self.create_quant_tensor(out)
                    if not self.training and self.cache_inference_quant_weight and self._cached_weight is None:
                        self._cached_weight = self.cache_class(
                            out.detach(),
                            metadata_only=self.cache_inference_quant_weight_metadata_only)
        else:  # quantization disabled
            out = self.apply_input_view(x)
        return out


class BiasQuantProxyFromInjectorBase(ParameterQuantProxyFromInjector, BiasQuantProxyProtocol, ABC):

    def __init__(self, quant_layer: nn.Module, quant_injector: Injector) -> None:
        super().__init__(quant_layer, quant_injector)
        self._cached_bias = None
        self.cache_inference_quant_bias = False
        self.cache_inference_quant_bias_metadata_only = False
        self.requires_input_scale = self.quant_injector.requires_input_scale

    @property
    def tracked_parameter_list(self):
        return [m.bias for m in self.tracked_module_list if m.bias is not None]

    def get_cached(self, attr):
        if self._cached_bias is None:
            if not is_dynamo_compiling():
                warn(
                    "No quant bias cache found, set cache_inference_quant_bias=True and run an "
                    "inference pass first")
            return None
        if self.training:
            warn("Cached quant bias scale is being used in training mode.")
        return getattr(self._cached_bias, attr)


class WeightQuantProxyFromInjector(WeightQuantProxyFromInjectorBase):

    def __init__(self, quant_layer: nn.Module, quant_injector: Injector) -> None:
        super().__init__(quant_layer, quant_injector)
        self.cache_class = _CachedIO

    @property
    def tracked_parameter_list(self):
        return [m.weight for m in self.tracked_module_list if m.weight is not None]

    @property
    def requires_quant_input(self):
        return False

    def scale(self):
        if not self.is_quant_enabled:
            return None
        scale = self.__call__(self.tracked_parameter_list[0]).scale
        return scale

    def zero_point(self):
        if not self.is_quant_enabled:
            return None
        zero_point = self.__call__(self.tracked_parameter_list[0]).zero_point
        return zero_point

    def bit_width(self):
        if not self.is_quant_enabled:
            return None
        bit_width = self.__call__(self.tracked_parameter_list[0]).bit_width
        return bit_width

    def create_quant_tensor(self, qt_args: Tuple[Any]) -> IntQuantTensor:
        return IntQuantTensor(*qt_args, self.is_signed, self.training)


class DecoupledWeightQuantProxyFromInjector(WeightQuantProxyFromInjector):

    def pre_scale(self):
        if not self.is_quant_enabled:
            return None
        output_tuple = self.tensor_quant(self.tracked_parameter_list[0])
        out, scale, zero_point, bit_width, pre_scale, pre_zero_point = output_tuple
        return pre_scale

    def pre_zero_point(self):
        if not self.is_quant_enabled:
            return None
        output_tuple = self.tensor_quant(self.tracked_parameter_list[0])
        out, scale, zero_point, bit_width, pre_scale, pre_zero_point = output_tuple
        return pre_zero_point

    def create_quant_tensor(self, qt_args: Tuple[Any]) -> IntQuantTensor:
        out, scale, zero_point, bit_width, pre_scale, pre_zero_point = qt_args
        return IntQuantTensor(out, scale, zero_point, bit_width, self.is_signed, self.training)


class DecoupledWeightQuantWithInputProxyFromInjector(DecoupledWeightQuantProxyFromInjector):

    def __init__(self, quant_layer: nn.Module, quant_injector: Injector) -> None:
        super().__init__(quant_layer, quant_injector)
        # Necessary for export
        self._cached_act = None
        self.cache_inference_quant_act = False
        self.cache_quant_io_metadata_only = True

    @property
    def requires_quant_input(self):
        return True

    def pre_scale(self):
        raise NotImplementedError

    def pre_zero_point(self):
        raise NotImplementedError

    def forward(
        self,
        x: torch.Tensor,
        quant_input: Optional[Union[Tensor,
                                    IntQuantTensor]] = None) -> Union[Tensor, IntQuantTensor]:
        if isinstance(quant_input,
                      IntQuantTensor) and not self.training and self.cache_inference_quant_act:
            cached_inp = _CachedIO(quant_input.detach(), self.cache_quant_io_metadata_only)
            self._cached_act = cached_inp

        if self.is_quant_enabled:
            if quant_input is None:
                assert self._cached_act is not None, "No cached quant input found. Enable caching and perform a forward pass"
                quant_input = self._cached_act
            else:
                assert isinstance(quant_input, IntQuantTensor), "Input must be quantized"

            input_bit_width = quant_input.bit_width
            input_is_signed = quant_input.signed

            impl = self.export_handler if self.export_mode else self.tensor_quant
            out, scale, zero_point, bit_width, pre_scale, pre_zero_point = impl(x, input_bit_width, input_is_signed)
            return IntQuantTensor(out, scale, zero_point, bit_width, self.is_signed, self.training)
        else:  # quantization disabled
            return x


class BiasQuantProxyFromInjector(BiasQuantProxyFromInjectorBase):

    def scale(self):
        if not self.is_quant_enabled:
            return None
        if self.requires_input_scale and self.is_quant_enabled:
            cache = self.get_cached('scale')
            return cache
        zhs = self._zero_hw_sentinel()
        scale = self.__call__(self.tracked_parameter_list[0], zhs).scale
        return scale

    def zero_point(self):
        if not self.is_quant_enabled:
            return None
        zhs = self._zero_hw_sentinel()
        zero_point = self.__call__(self.tracked_parameter_list[0], zhs).zero_point
        return zero_point

    def bit_width(self):
        if not self.is_quant_enabled:
            return None
        zhs = self._zero_hw_sentinel()
        bit_width = self.__call__(self.tracked_parameter_list[0], zhs).bit_width
        return bit_width

    def quant_output_scale_impl(
            self, input: IntQuantTensor, weight: IntQuantTensor, module: torch.nn.Module) -> Tensor:
        channel_dim = -1 if isinstance(module, torch.nn.Linear) else 1
        output_scale_shape = compute_channel_view_shape(input, channel_dim=channel_dim)
        output_scale = weight.scale.view(output_scale_shape)
        output_scale = output_scale * input.scale.view(output_scale_shape)
        return output_scale

    def compute_bias_scale(
            self,
            input: Optional[Union[Tensor, IntQuantTensor]],
            weight: Optional[Union[Tensor, IntQuantTensor]]) -> Optional[Tensor]:
        if not self.requires_input_scale and self.is_quant_enabled:
            return None
        if not isinstance(input, IntQuantTensor) or not isinstance(weight, IntQuantTensor):
            return None
        if len(self.tracked_module_list) > 1:
            if not all(
                [type[self.tracked_module_list[0]] == type[x] for x in self.tracked_module_list]):
                raise RuntimeError(
                    "Bias quantizer shared across different type of layers with external scale is not supported."
                )
        scale = self.quant_output_scale_impl(input, weight, self.tracked_module_list[0])
        return scale

    def forward(
            self,
            x: Tensor,
            input: Optional[Union[Tensor, IntQuantTensor]] = None,
            weight: Optional[Union[Tensor,
                                   IntQuantTensor]] = None) -> Union[Tensor, IntQuantTensor]:
        out = x
        input_scale = self.compute_bias_scale(input, weight)
        if self.is_quant_enabled:
            impl = self.export_handler if self.export_mode else self.tensor_quant
            if self.requires_input_scale and input_scale is None and self.is_quant_enabled:
                input_scale = self.scale()
                if input_scale is None:
                    raise RuntimeError("Input scale required")
            elif self.requires_input_scale and input_scale is not None and self.is_quant_enabled:
                input_scale = input_scale.view(-1)

            if self.requires_input_scale and self.is_quant_enabled:
                out, out_scale, out_zp, out_bit_width = impl(x, input_scale)
            else:
                out, out_scale, out_zp, out_bit_width = impl(x)
            if not is_dynamo_compiling():
                out = IntQuantTensor(
                    out, out_scale, out_zp, out_bit_width, self.is_signed, self.training)
                if not self.training and self.cache_inference_quant_bias:
                    cached_bias = _CachedIO(
                        out.detach(), metadata_only=self.cache_inference_quant_bias_metadata_only)
                    self._cached_bias = cached_bias
        else:
            out = x
        return out



from typing import NamedTuple
from brevitas.quant_tensor.int_quant_tensor import IntQuantTensor
import torch

# Remove the bases check
def is_namedtuple_cls_patched(cls): 
    """Test if an object is a namedtuple or a (torch.return_types|torch.autograd.forward_ad).* quasi-namedtuple""" 
    try: 
        if issubclass(cls, tuple): 
            module = getattr(cls, "__module__", None) 
            return module in ("torch.return_types", "torch.autograd.forward_ad") or ( 
                hasattr(cls, "_make") and hasattr(cls, "_fields") 
            )
    except TypeError:
        pass
    return False

# Not the prettiest but it does the job
torch._dynamo.utils.is_namedtuple_cls.__code__ = is_namedtuple_cls_patched.__code__

torch._dynamo.allow_in_graph
def create_nt(a1, a2):
    return BaseDtype(a1, a2)

class BaseDtype(NamedTuple):
    a: torch.Tensor
    b: torch.Tensor

    def __add__(self, other):
        return create_nt(self.a + other.a, self.b + other.b)
    @classmethod
    def __torch_function__(cls, func, types, args=(), kwargs=None):
        args = (args[0].a)
        return func(args)

class MyDType(BaseDtype):

    def __new__(cls, a, b):
        # This is needed to do type checking on init variables
        return super().__new__(cls, a, b)

    def __add__(self, other):
        return MyDType(self.a + other.a, self.b)

    @classmethod
    def __torch_function__(cls, func, types, args=(), kwargs=None):
        args = (args[0].a)
        return func(args)

# This is required to create MyDType within the forward pass
# Removing __new__ and inheriting directly from NamedTuple seems to work but it does not fit my use case
# torch._dynamo.allow_in_graph(MyDType)



class Model(torch.nn.Module):
  def __init__(self):
    super().__init__()
    
  def forward(self, a1, a2):
    inp1 = create_nt(a1, a2)
    inp2 = create_nt(a1, a2)
    return torch.nn.functional.relu(inp1 + inp2)


model = Model()
non_compiled_output = model(torch.tensor(3.), torch.tensor(0.))
print(f"Correct output: {non_compiled_output}")
model = torch.compile(model, fullgraph=True)
compiled_output = model(torch.tensor(3.), torch.tensor(0.))
# Correct output: MyDType(a=tensor(6.), b=tensor(0.))
print(f"Less correct output: {compiled_output}")
# Less correct output: (tensor(3.), tensor(0.), tensor(3.), tensor(0.))


# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from torch import nn

from brevitas.fx.brevitas_tracer import symbolic_trace
from brevitas.graph.per_input import AdaptiveAvgPoolToAvgPool
from brevitas.graph.quantize import preprocess_for_quantize
from brevitas.graph.quantize import quantize
from brevitas.graph.quantize import UNSIGNED_ACT_TUPLE
from brevitas.graph.standardize import MeanMethodToAdaptiveAvgPool2d
import brevitas.nn as qnn
from brevitas.quant import Int8ActPerTensorFixedPoint
from brevitas.quant import Int8WeightPerTensorFixedPoint
from brevitas.quant import Int16Bias
from brevitas.quant import Uint8ActPerTensorFixedPoint
from brevitas.quant import Uint8ActPerTensorFixedPointMaxInit
from brevitas.quant.fixed_point import Int8ActPerTensorFixedPointMinMaxInit

FLEXML_COMPUTE_LAYER_MAP = {
    nn.AvgPool2d: (qnn.flexml.FlexMLQuantAvgPool2d, {
        'return_quant_tensor': True}),
    nn.MultiheadAttention: (
        qnn.QuantMultiheadAttention,
        {
            'in_proj_weight_quant': Int8WeightPerTensorFixedPoint,
            'in_proj_bias_quant': Int16Bias,
            'attn_output_weights_quant': Uint8ActPerTensorFixedPoint,
            'q_scaled_quant': Int8ActPerTensorFixedPoint,
            'k_transposed_quant': Int8ActPerTensorFixedPoint,
            'v_quant': Int8ActPerTensorFixedPoint,
            'out_proj_input_quant': Int8ActPerTensorFixedPoint,
            'out_proj_weight_quant': Int8WeightPerTensorFixedPoint,
            'out_proj_bias_quant': Int16Bias,
            'return_quant_tensor': True}),
    nn.Conv1d: (
        qnn.QuantConv1d,
        {
            'weight_quant': Int8WeightPerTensorFixedPoint,
            'bias_quant': Int16Bias,
            'return_quant_tensor': True}),
    nn.Conv2d: (
        qnn.QuantConv2d,
        {
            'weight_quant': Int8WeightPerTensorFixedPoint,
            'bias_quant': Int16Bias,
            'return_quant_tensor': True}),
    nn.Conv3d: (
        qnn.QuantConv3d,
        {
            'weight_quant': Int8WeightPerTensorFixedPoint,
            'bias_quant': Int16Bias,
            'return_quant_tensor': True}),
    nn.ConvTranspose1d: (
        qnn.QuantConvTranspose1d,
        {
            'weight_quant': Int8WeightPerTensorFixedPoint,
            'bias_quant': Int16Bias,
            'return_quant_tensor': True}),
    nn.ConvTranspose2d: (
        qnn.QuantConvTranspose2d,
        {
            'weight_quant': Int8WeightPerTensorFixedPoint,
            'bias_quant': Int16Bias,
            'return_quant_tensor': True}),
    nn.ConvTranspose3d: (
        qnn.QuantConvTranspose3d,
        {
            'weight_quant': Int8WeightPerTensorFixedPoint,
            'bias_quant': Int16Bias,
            'return_quant_tensor': True}),
    nn.BatchNorm1d: (
        qnn.BatchNorm1dToQuantScaleBias,
        {
            'weight_quant': Int8WeightPerTensorFixedPoint,
            'bias_quant': Int16Bias,
            'return_quant_tensor': True}),
    nn.BatchNorm2d: (
        qnn.BatchNorm2dToQuantScaleBias,
        {
            'weight_quant': Int8WeightPerTensorFixedPoint,
            'bias_quant': Int16Bias,
            'return_quant_tensor': True}),
    nn.Linear: (
        qnn.QuantLinear,
        {
            'weight_quant': Int8WeightPerTensorFixedPoint,
            'bias_quant': Int16Bias,
            'return_quant_tensor': True})}

FLEXML_QUANT_ACT_MAP = {
    nn.ReLU:
        (qnn.QuantReLU, {
            'act_quant': Uint8ActPerTensorFixedPoint, 'return_quant_tensor': True}),
    nn.ReLU6: (
        qnn.QuantReLU, {
            'act_quant': Uint8ActPerTensorFixedPointMaxInit,
            'max_val': 6.,
            'return_quant_tensor': True}),
    nn.LeakyReLU: (
        qnn.flexml.FlexMLQuantLeakyReLU,
        {
            'alpha_quant':
                qnn.QuantIdentity(Uint8ActPerTensorFixedPoint, bit_width=16),
            'input_quant':
                qnn.QuantIdentity(
                    Int8ActPerTensorFixedPoint, bit_width=16, scaling_stats_momentum=None),
            'output_quant':
                qnn.QuantIdentity(Int8ActPerTensorFixedPoint, return_quant_tensor=True)}),
    nn.Hardtanh: (
        qnn.QuantHardTanh,
        {
            'act_quant': Int8ActPerTensorFixedPointMinMaxInit,
            'max_val': lambda module: module.max_val,
            'min_val': lambda module: module.min_val,
            'return_quant_tensor': True}),
    nn.Sigmoid: (
        qnn.QuantSigmoid, {
            'act_quant': Uint8ActPerTensorFixedPoint,
            'return_quant_tensor': True,}),}

FLEXML_QUANT_IDENTITY_MAP = {
    'signed':
        (qnn.QuantIdentity, {
            'act_quant': Int8ActPerTensorFixedPoint, 'return_quant_tensor': True}),
    'unsigned': (
        qnn.QuantIdentity, {
            'act_quant': Uint8ActPerTensorFixedPoint, 'return_quant_tensor': True}),}


def preprocess_for_flexml_quantize(
        model,
        *model_args,
        trace_model=True,
        relu6_to_relu=True,
        equalize_iters=0,
        equalize_merge_bias=True,
        merge_bn=True,
        equalize_bias_shrinkage='vaiq',
        equalize_scale_computation='maxabs',
        **model_kwargs):
    training_state = model.training
    model.eval()

    if trace_model:
        model = symbolic_trace(model)
    model = MeanMethodToAdaptiveAvgPool2d().apply(model)
    model = AdaptiveAvgPoolToAvgPool().apply(model, *model_args, **model_kwargs)
    model = preprocess_for_quantize(
        model,
        False,
        relu6_to_relu,
        equalize_iters,
        equalize_merge_bias,
        merge_bn,
        equalize_bias_shrinkage,
        equalize_scale_computation)
    model.train(training_state)
    return model


def quantize_flexml(
        graph_model,
        quant_identity_map=FLEXML_QUANT_IDENTITY_MAP,
        compute_layer_map=FLEXML_COMPUTE_LAYER_MAP,
        quant_act_map=FLEXML_QUANT_ACT_MAP,
        unsigned_act_tuple=UNSIGNED_ACT_TUPLE,
        requantize_layer_handler_output=True):
    return quantize(
        graph_model,
        quant_identity_map=quant_identity_map,
        compute_layer_map=compute_layer_map,
        quant_act_map=quant_act_map,
        unsigned_act_tuple=unsigned_act_tuple,
        requantize_layer_handler_output=requantize_layer_handler_output)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

# Part of this code has been re-adapted from https://github.com/yhhhli/BRECQ
# under the following LICENSE:

# MIT License

# Copyright (c) 2021 Yuhang Li

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
import torch
import torch.nn.functional as F

from brevitas import config
from brevitas.core.function_wrapper.learned_round import LearnedRoundSte
from brevitas.graph.calibrate import DisableEnableQuantization
from brevitas.inject.enum import FloatToIntImplType
from brevitas.inject.enum import LearnedRoundImplType
from brevitas.nn.quant_layer import QuantWeightBiasInputOutputLayer as QuantWBIOL
from brevitas.quant_tensor import QuantTensor

config.IGNORE_MISSING_KEYS = True


class StopFwdException(Exception):
    """Used to throw and catch an exception to stop traversing the graph."""
    pass


class DataSaverHook:

    def __init__(self, store_output: False):
        self.store_output = store_output
        self.input_store = None
        self.output_store = None

    def __call__(self, module, input_batch, output_batch):
        input_batch = input_batch[0]
        if isinstance(input_batch, QuantTensor):
            input_batch = input_batch.value

        if hasattr(input_batch, 'names') and 'N' in input_batch.names:
            batch_dim = input_batch.names.index('N')

            input_batch.rename_(None)
            input_batch = input_batch.transpose(0, batch_dim)
            if self.store_output:
                output_batch.rename_(None)
                output_batch = output_batch.transpose(0, batch_dim)

        if self.store_output:
            self.output_store = output_batch
        self.input_store = input_batch
        raise StopFwdException


class LinearTempDecay:

    def __init__(self, t_max: int, rel_start_decay: float = 0.2, start_b: int = 10, end_b: int = 2):
        self.t_max = t_max
        self.start_decay = rel_start_decay * t_max
        self.start_b = start_b
        self.end_b = end_b

    def __call__(self, t):
        if t < self.start_decay:
            return self.start_b
        else:
            rel_t = (t - self.start_decay) / (self.t_max - self.start_decay)
            return self.end_b + (self.start_b - self.end_b) * max(0.0, (1 - rel_t))


class Loss:

    def __init__(
            self,
            module,
            learned_round_module,
            weight=0.01,
            max_count=1000,
            b_range=(20, 2),
            warmup=0.2,
            decay_start=0.0):
        self.weight = weight
        self.module = module
        self.loss_start = max_count * warmup
        self.temp_decay = LinearTempDecay(
            max_count,
            start_b=b_range[0],
            end_b=b_range[1],
            rel_start_decay=warmup + (1.0 - warmup) * decay_start)
        self.iter = 0
        self.learned_round_module = learned_round_module

    def __call__(self, pred, tgt):
        self.iter += 1

        rec_loss = F.mse_loss(pred, tgt, reduction='none').sum(1).mean()

        if self.iter < self.loss_start:
            b = self.temp_decay(self.iter)
            round_loss = 0
        else:  # 1 - |(h-0.5)*2|**b
            b = self.temp_decay(self.iter)
            round_vals = self.learned_round_module.p_forward()
            round_loss = self.weight * (1 - ((round_vals - 0.5).abs() * 2).pow(b)).sum()

        total_loss = rec_loss + round_loss
        return total_loss, rec_loss, round_loss, b


def find_learned_round_module(module):
    for submodule in module.modules():
        if isinstance(submodule, LearnedRoundSte):
            return submodule
    return False


def insert_learned_round_quantizer(layer, learned_round_zeta=1.1, learned_round_gamma=-0.1):
    if isinstance(layer, QuantWBIOL):
        if not find_learned_round_module(layer):
            floor_weight = torch.floor(layer.weight.data / layer.quant_weight().scale)
            delta = (layer.weight.data / layer.quant_weight().scale) - floor_weight
            value = -torch.log((learned_round_zeta - learned_round_gamma) /
                               (delta - learned_round_gamma) - 1)
            layer.weight_quant.quant_injector = layer.weight_quant.quant_injector.let(
                float_to_int_impl_type=FloatToIntImplType.LEARNED_ROUND,
                learned_round_impl_type=LearnedRoundImplType.HARD_SIGMOID,
                learned_round_gamma=learned_round_gamma,
                learned_round_zeta=learned_round_zeta,
                learned_round_init=value)
            layer.weight_quant.init_tensor_quant(preserve_state_dict=True)


def split_layers(model, layers):
    for module in model.children():
        if isinstance(module, QuantWBIOL):
            layers.append(module)
        else:
            split_layers(module, layers)


def learned_round_iterator(layers, iters=1000):
    for layer in layers:
        insert_learned_round_quantizer(layer)

        for p in layer.parameters():
            p.requires_grad = False

        learned_round_module = find_learned_round_module(layer)
        learned_round_module.value.requires_grad = True
        layer_loss = Loss(module=layer, learned_round_module=learned_round_module, max_count=iters)
        yield layer, layer_loss, learned_round_module
        layer.eval()


def save_inp_out_data(
        model,
        module,
        dataloader: torch.utils.data.DataLoader,
        store_inp=False,
        store_out=False,
        keep_gpu: bool = True,
        disable_quant=False):
    if disable_quant:
        disable_quant_class = DisableEnableQuantization()
        disable_quant_class.disable_act_quantization(model, False)
        disable_quant_class.disable_param_quantization(model, False)
    device = next(model.parameters()).device
    data_saver = DataSaverHook(store_output=store_out)
    handle = module.register_forward_hook(data_saver)
    cached = [[], []]
    with torch.no_grad():
        for img, t in dataloader:
            try:
                _ = model(img.to(device))
            except StopFwdException:
                pass
            if store_inp:
                if keep_gpu:
                    cached[0].append(data_saver.input_store.detach())
                else:
                    cached[0].append(data_saver.input_store.detach().cpu())
            if store_out:
                if keep_gpu:
                    cached[1].append(data_saver.output_store.detach())
                else:
                    cached[1].append(data_saver.output_store.detach().cpu())
    if store_inp:
        cached[0] = torch.cat([x for x in cached[0]])
    if store_out:
        cached[1] = torch.cat([x for x in cached[1]])
    handle.remove()
    if disable_quant:
        disable_quant_class.enable_act_quantization(model, False)
        disable_quant_class.enable_param_quantization(model, False)
    return cached



from abc import ABC
from typing import Any, Optional, Tuple, Union

from torch import Tensor
import torch.nn as nn

from brevitas.core.function_wrapper.misc import Identity
from brevitas.inject import BaseInjector as Injector
from brevitas.proxy.parameter_quant import BiasQuantProxyFromInjectorBase
from brevitas.proxy.parameter_quant import WeightQuantProxyFromInjectorBase
from brevitas.quant_tensor import FloatQuantTensor
from brevitas.utils.quant_utils import _CachedIOFloat


class WeightFloatQuantProxyFromInjectorBase(WeightQuantProxyFromInjectorBase, ABC):

    def scale(self):
        if not self.is_quant_enabled:
            return None
        scale = self.__call__(self.tracked_parameter_list[0]).scale
        return scale

    def zero_point(self):
        if not self.is_quant_enabled:
            return None
        zero_point = self.__call__(self.tracked_parameter_list[0]).zero_point
        return zero_point

    def exponent_bit_width(self):
        if not self.is_quant_enabled:
            return None
        exponent_bit_width = self.__call__(self.tracked_parameter_list[0]).exponent_bit_width
        return exponent_bit_width

    def mantissa_bit_width(self):
        if not self.is_quant_enabled:
            return None
        mantissa_bit_width = self.__call__(self.tracked_parameter_list[0]).mantissa_bit_width
        return mantissa_bit_width

    def exponent_bias(self):
        if not self.is_quant_enabled:
            return None
        exponent_bias = self.__call__(self.tracked_parameter_list[0]).exponent_bias
        return exponent_bias

    def is_saturating(self):
        if not self.is_quant_enabled:
            return None
        saturating = self.__call__(self.tracked_parameter_list[0]).saturating
        return saturating

    def inf_values(self):
        if not self.is_quant_enabled:
            return None
        inf_values = self.__call__(self.tracked_parameter_list[0]).inf_values
        return inf_values

    def nan_values(self):
        if not self.is_quant_enabled:
            return None
        nan_values = self.__call__(self.tracked_parameter_list[0]).nan_values
        return nan_values

    @property
    def is_ocp(self):
        is_e4m3 = self.mantissa_bit_width() == 3 and self.exponent_bit_width() == 4
        is_ocp_e4m3 = is_e4m3 and self.inf_values() is None and self.nan_values() == (('111',))

        is_e5m2 = self.mantissa_bit_width() == 2 and self.exponent_bit_width() == 5
        is_ocp_e5m2 = is_e5m2 and self.inf_values() == (
            ('00',)) and self.nan_values() == ('01', '11', '10')

        return is_ocp_e4m3 or is_ocp_e5m2

    @property
    def is_fnuz(self):
        is_e4m3 = self.mantissa_bit_width() == 3 and self.exponent_bit_width() == 4
        is_fnuz_e4m3 = is_e4m3 and self.inf_values() is None and self.nan_values(
        ) is None and self.exponent_bias() == 8

        is_e5m2 = self.mantissa_bit_width() == 2 and self.exponent_bit_width() == 5
        is_fnuz_e5m2 = is_e5m2 and self.inf_values() is None and self.nan_values(
        ) is None and self.exponent_bias() == 16
        return is_fnuz_e4m3 or is_fnuz_e5m2

    @property
    def input_view_impl(self):
        if self.tensor_quant is not None:
            return self.tensor_quant.input_view_impl
        else:
            return Identity()


class WeightFloatQuantProxyFromInjector(WeightFloatQuantProxyFromInjectorBase):

    def __init__(self, quant_layer: nn.Module, quant_injector: Injector) -> None:
        super().__init__(quant_layer, quant_injector)
        self.cache_class = _CachedIOFloat

    def create_quant_tensor(self, qt_args: Tuple[Any]) -> FloatQuantTensor:
        out, scale, zero_point, exponent_bit_width, mantissa_bit_width, exponent_bias, saturating, inf_values, nan_values = qt_args
        return FloatQuantTensor(
            out,
            scale,
            zero_point,
            exponent_bit_width,
            mantissa_bit_width,
            exponent_bias,
            saturating,
            inf_values,
            nan_values,
            self.is_signed,
            self.training)


class BiasFloatQuantProxyFromInjector(BiasQuantProxyFromInjectorBase):

    def scale(self):
        if not self.is_quant_enabled:
            return None
        if self.requires_input_scale:
            cache = self.get_cached('scale')
            return cache
        zhs = self._zero_hw_sentinel()
        scale = self.__call__(self.tracked_parameter_list[0], zhs).scale
        return scale

    def zero_point(self):
        if not self.is_quant_enabled:
            return None
        zhs = self._zero_hw_sentinel()
        zero_point = self.__call__(self.tracked_parameter_list[0], zhs).zero_point
        return zero_point

    def exponent_bit_width(self):
        if not self.is_quant_enabled:
            return None
        exponent_bit_width = self.__call__(self.tracked_parameter_list[0]).exponent_bit_width
        return exponent_bit_width

    def mantissa_bit_width(self):
        if not self.is_quant_enabled:
            return None
        mantissa_bit_width = self.__call__(self.tracked_parameter_list[0]).mantissa_bit_width
        return mantissa_bit_width

    def exponent_bias(self):
        if not self.is_quant_enabled:
            return None
        exponent_bias = self.__call__(self.tracked_parameter_list[0]).exponent_bias
        return exponent_bias

    def is_saturating(self):
        if not self.is_quant_enabled:
            return None
        saturating = self.__call__(self.tracked_parameter_list[0]).saturating
        return saturating

    def inf_values(self):
        if not self.is_quant_enabled:
            return None
        inf_values = self.__call__(self.tracked_parameter_list[0]).inf_values
        return inf_values

    def nan_values(self):
        if not self.is_quant_enabled:
            return None
        nan_values = self.__call__(self.tracked_parameter_list[0]).nan_values
        return nan_values

    def forward(self,
                x: Tensor,
                input_scale: Optional[Tensor] = None) -> Union[Tensor, FloatQuantTensor]:
        out = x
        if self.is_quant_enabled:
            impl = self.export_handler if self.export_mode else self.tensor_quant
            if self.requires_input_scale and input_scale is None:
                input_scale = self.scale()
                if input_scale is None:
                    raise RuntimeError("Input scale required")

            if self.requires_input_scale:
                input_scale = input_scale.view(-1)
                out, scale, zero_point, exponent_bit_width, mantissa_bit_width, exponent_bias, saturating, inf_values, nan_values = impl(x, input_scale)
            else:
                out, scale, zero_point, exponent_bit_width, mantissa_bit_width, exponent_bias, saturating, inf_values, nan_values = impl(x)

            out = FloatQuantTensor(
                out,
                scale,
                zero_point,
                exponent_bit_width,
                mantissa_bit_width,
                exponent_bias,
                saturating,
                inf_values,
                nan_values,
                self.is_signed,
                self.training)
        else:
            out = x
        if isinstance(out,
                      FloatQuantTensor) and not self.training and self.cache_inference_quant_bias:
            cached_bias = _CachedIOFloat(out.detach(), metadata_only=False)
            self._cached_bias = cached_bias
        return out



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import argparse
from functools import partial
import os
import random
import warnings

import numpy as np
import torch
import torch.backends.cudnn as cudnn
import torch.nn.parallel
import torch.optim
import torch.utils.data
import torch.utils.data.distributed
import torchvision

from brevitas.export import export_onnx_qcdq
from brevitas.export import export_torch_qcdq
from brevitas.export.inference import quant_inference_mode
from brevitas.graph.quantize import preprocess_for_quantize
from brevitas.graph.target.flexml import preprocess_for_flexml_quantize
from brevitas_examples.imagenet_classification.ptq.ptq_common import apply_act_equalization
from brevitas_examples.imagenet_classification.ptq.ptq_common import apply_bias_correction
from brevitas_examples.imagenet_classification.ptq.ptq_common import apply_gpfq
from brevitas_examples.imagenet_classification.ptq.ptq_common import apply_gptq
from brevitas_examples.imagenet_classification.ptq.ptq_common import apply_learned_round_learning
from brevitas_examples.imagenet_classification.ptq.ptq_common import calibrate
from brevitas_examples.imagenet_classification.ptq.ptq_common import calibrate_bn
from brevitas_examples.imagenet_classification.ptq.ptq_common import quantize_model
from brevitas_examples.imagenet_classification.ptq.utils import add_bool_arg
from brevitas_examples.imagenet_classification.ptq.utils import get_model_config
from brevitas_examples.imagenet_classification.ptq.utils import get_torchvision_model
from brevitas_examples.imagenet_classification.utils import generate_dataloader
from brevitas_examples.imagenet_classification.utils import SEED
from brevitas_examples.imagenet_classification.utils import validate

# Ignore warnings about __torch_function__
warnings.filterwarnings("ignore")


def parse_type(v, default_type):
    if v == 'None':
        return None
    else:
        return default_type(v)


model_names = sorted(
    name for name in torchvision.models.__dict__ if name.islower() and not name.startswith("__") and
    callable(torchvision.models.__dict__[name]) and not name.startswith("get_"))

parser = argparse.ArgumentParser(description='PyTorch ImageNet PTQ Validation')
parser.add_argument(
    '--calibration-dir',
    required=True,
    help='Path to folder containing Imagenet calibration folder')
parser.add_argument(
    '--validation-dir', required=True, help='Path to folder containing Imagenet validation folder')
parser.add_argument(
    '--workers', default=8, type=int, help='Number of data loading workers (default: 8)')
parser.add_argument(
    '--batch-size-calibration',
    default=64,
    type=int,
    help='Minibatch size for calibration (default: 64)')
parser.add_argument(
    '--batch-size-validation',
    default=256,
    type=int,
    help='Minibatch size for validation (default: 256)')
parser.add_argument(
    '--export-dir', default='.', type=str, help='Directory where to store the exported models')
parser.add_argument('--gpu', default=None, type=int, help='GPU id to use (default: None)')
parser.add_argument(
    '--calibration-samples', default=1000, type=int, help='Calibration size (default: 1000)')
parser.add_argument(
    '--model-name',
    default='resnet18',
    metavar='ARCH',
    choices=model_names,
    help='model architecture: ' + ' | '.join(model_names) + ' (default: resnet18)')
parser.add_argument(
    '--dtype', default='float', choices=['float', 'bfloat16'], help='Data type to use')
parser.add_argument(
    '--target-backend',
    default='fx',
    choices=['fx', 'layerwise', 'flexml'],
    help='Backend to target for quantization (default: fx)')
parser.add_argument(
    '--scale-factor-type',
    default='float_scale',
    choices=['float_scale', 'po2_scale'],
    help='Type for scale factors (default: float_scale)')
parser.add_argument(
    '--act-bit-width', default=8, type=int, help='Activations bit width (default: 8)')
parser.add_argument(
    '--weight-bit-width', default=8, type=int, help='Weights bit width (default: 8)')
parser.add_argument(
    '--layerwise-first-last-bit-width',
    default=8,
    type=int,
    help='Input and weights bit width for first and last layer w/ layerwise backend (default: 8)')
parser.add_argument(
    '--bias-bit-width',
    default=32,
    type=partial(parse_type, default_type=int),
    choices=[32, 16, None],
    help='Bias bit width (default: 32)')
parser.add_argument(
    '--act-quant-type',
    default='sym',
    choices=['sym', 'asym'],
    help='Activation quantization type (default: sym)')
parser.add_argument(
    '--weight-quant-type',
    default='sym',
    choices=['sym', 'asym'],
    help='Weight quantization type (default: sym)')
parser.add_argument(
    '--weight-quant-granularity',
    default='per_tensor',
    choices=['per_tensor', 'per_channel', 'per_group'],
    help='Weight quantization type (default: per_tensor)')
parser.add_argument(
    '--act-quant-granularity',
    default='per_tensor',
    choices=['per_tensor', 'per_group'],
    help='Activation quantization type (default: per_tensor)')
parser.add_argument(
    '--weight-quant-calibration-type',
    default='stats',
    choices=['stats', 'mse', 'hqo'],
    help='Weight quantization calibration type (default: stats)')
parser.add_argument(
    '--act-equalization',
    default=None,
    choices=['fx', 'layerwise', None],
    help='Activation equalization type (default: None)')
parser.add_argument(
    '--act-quant-calibration-type',
    default='stats',
    choices=['stats', 'mse'],
    help='Activation quantization calibration type (default: stats)')
parser.add_argument(
    '--act-scale-computation-type',
    default='static',
    choices=['static', 'dynamic'],
    help='Activation quantization scale computation type (default: static)')
parser.add_argument(
    '--graph-eq-iterations',
    default=20,
    type=int,
    help='Numbers of iterations for graph equalization (default: 20)')
parser.add_argument(
    '--learned-round-iters',
    default=1000,
    type=int,
    help='Numbers of iterations for learned round for each layer (default: 1000)')
parser.add_argument(
    '--learned-round-lr',
    default=1e-3,
    type=float,
    help='Learning rate for learned round (default: 1e-3)')
parser.add_argument(
    '--act-quant-percentile',
    default=99.999,
    type=float,
    help='Percentile to use for stats of activation quantization (default: 99.999)')
parser.add_argument(
    '--export-onnx-qcdq', action='store_true', help='If true, export the model in onnx qcdq format')
parser.add_argument(
    '--export-torch-qcdq',
    action='store_true',
    help='If true, export the model in torch qcdq format')

add_bool_arg(
    parser, 'bias-corr', default=True, help='Bias correction after calibration (default: enabled)')
add_bool_arg(
    parser,
    'graph-eq-merge-bias',
    default=True,
    help='Merge bias when performing graph equalization (default: enabled)')
add_bool_arg(
    parser,
    'weight-narrow-range',
    default=False,
    help='Narrow range for weight quantization (default: disabled)')
parser.add_argument('--gpfq-p', default=1.0, type=float, help='P parameter for GPFQ (default: 1.0)')
parser.add_argument(
    '--quant-format',
    default='int',
    choices=['int', 'float', 'float_ocp'],
    help='Quantization format to use for weights and activations (default: int)')
parser.add_argument(
    '--layerwise-first-last-mantissa-bit-width',
    default=4,
    type=int,
    help=
    'Mantissa bit width used with float layerwise quantization for first and last layer (default: 4)'
)
parser.add_argument(
    '--layerwise-first-last-exponent-bit-width',
    default=3,
    type=int,
    help=
    'Exponent bit width used with float layerwise quantization for first and last layer (default: 3)'
)
parser.add_argument(
    '--weight-mantissa-bit-width',
    default=4,
    type=int,
    help='Mantissa bit width used with float quantization for weights (default: 4)')
parser.add_argument(
    '--weight-exponent-bit-width',
    default=3,
    type=int,
    help='Exponent bit width used with float quantization for weights (default: 3)')
parser.add_argument(
    '--act-mantissa-bit-width',
    default=4,
    type=int,
    help='Mantissa bit width used with float quantization for activations (default: 4)')
parser.add_argument(
    '--act-exponent-bit-width',
    default=3,
    type=int,
    help='Exponent bit width used with float quantization for activations (default: 3)')
parser.add_argument(
    '--accumulator-bit-width',
    default=None,
    type=int,
    help='Accumulator Bit Width for GPFA2Q (default: None)')
parser.add_argument('--onnx-opset-version', default=None, type=int, help='ONNX opset version')
parser.add_argument(
    '--channel-splitting-ratio',
    default=0.0,
    type=float,
    help=
    'Split Ratio for Channel Splitting. When set to 0.0, Channel Splitting will not be applied. (default: 0.0)'
)
parser.add_argument(
    '--compression-rate',
    default=0.0,
    type=float,
    help='Specify compression rate < 1.0 for random projection. Default is 0.0 and does not use RP.'
)
add_bool_arg(parser, 'gptq', default=False, help='GPTQ (default: disabled)')
add_bool_arg(parser, 'gpfq', default=False, help='GPFQ (default: disabled)')
add_bool_arg(parser, 'gpfa2q', default=False, help='GPFA2Q (default: disabled)')
add_bool_arg(
    parser, 'gpxq-act-order', default=False, help='GPxQ Act order heuristic (default: disabled)')
add_bool_arg(parser, 'learned-round', default=False, help='Learned round (default: disabled)')
add_bool_arg(parser, 'calibrate-bn', default=False, help='Calibrate BN (default: disabled)')
add_bool_arg(
    parser,
    'channel-splitting-split-input',
    default=False,
    help='Input Channels Splitting for channel splitting (default: disabled)')
add_bool_arg(
    parser,
    'merge-bn',
    default=True,
    help='Merge BN layers before quantizing the model (default: enabled)')
add_bool_arg(
    parser,
    'uint_sym_act_for_unsigned_values',
    default=True,
    help='Use unsigned act quant when possible (default: enabled)')
add_bool_arg(parser, 'compile', default=False, help='Use torch.compile (default: disabled)')


def generate_ref_input(args, device, dtype):
    model_config = get_model_config(args.model_name)
    center_crop_shape = model_config['center_crop_shape']
    img_shape = center_crop_shape
    return torch.ones(1, 3, img_shape, img_shape, device=device, dtype=dtype)


def main():
    args = parser.parse_args()
    dtype = getattr(torch, args.dtype)

    random.seed(SEED)
    np.random.seed(SEED)
    torch.manual_seed(SEED)
    if args.act_quant_calibration_type == 'stats':
        act_quant_calib_config = str(args.act_quant_percentile) + 'stats'
    else:
        act_quant_calib_config = args.act_quant_calibration_type

    if args.act_bit_width == 0:
        args.act_bit_width = None

    config = (
        f"{args.model_name}_"
        f"{args.target_backend}_"
        f"{args.quant_format}_"
        f"{str(args.weight_mantissa_bit_width) + '_' if args.quant_format == 'float' else ''}"
        f"{str(args.weight_exponent_bit_width) + '_' if args.quant_format == 'float' else ''}"
        f"{str(args.act_mantissa_bit_width) + '_' if args.quant_format == 'float' else ''}"
        f"{str(args.act_exponent_bit_width) + '_' if args.quant_format == 'float' else ''}"
        f"{args.scale_factor_type}_"
        f"a{args.act_bit_width}"
        f"w{args.weight_bit_width}_"
        f"{'gptq_' if args.gptq else ''}"
        f"{'gpfq_' if args.gpfq else ''}"
        f"{'gpfa2q_' if args.gpfa2q else ''}"
        f"{'gpxq_act_order_' if args.gpxq_act_order else ''}"
        f"{'learned_round_' if args.learned_round else ''}"
        f"{'weight_narrow_range_' if args.weight_narrow_range else ''}"
        f"{args.bias_bit_width}bias_"
        f"{args.weight_quant_granularity}_"
        f"{args.act_quant_type}_"
        f"{'bc_' if args.bias_corr else ''}"
        f"{args.graph_eq_iterations}geiters_"
        f"{'mb_' if args.graph_eq_merge_bias else ''}"
        f"{act_quant_calib_config}_"
        f"{args.weight_quant_calibration_type}_"
        f"{'bnc_' if args.calibrate_bn else ''}"
        f"{'channel_splitting' if args.channel_splitting_ratio else ''}")

    print(
        f"Model: {args.model_name} - "
        f"Target backend: {args.target_backend} - "
        f"Quantization type: {args.scale_factor_type} - "
        f"Activation bit width: {args.act_bit_width} - "
        f"Weight bit width: {args.weight_bit_width} - "
        f"GPTQ: {args.gptq} - "
        f"GPFQ: {args.gpfq} - "
        f"GPFA2Q: {args.gpfa2q} - "
        f"GPFQ P: {args.gpfq_p} - "
        f"GPxQ Act Order: {args.gpxq_act_order} - "
        f"GPFA2Q Accumulator Bit Width: {args.accumulator_bit_width} - "
        f"Learned Round: {args.learned_round} - "
        f"Weight narrow range: {args.weight_narrow_range} - "
        f"Bias bit width: {args.bias_bit_width} - "
        f"Weight scale factors type: {args.weight_quant_granularity} - "
        f"Activation quant type: {args.act_quant_type} - "
        f"Bias Correction Enabled: {args.bias_corr} - "
        f"Iterations for graph equalization: {args.graph_eq_iterations} - "
        f"Merge bias in graph equalization: {args.graph_eq_merge_bias} - "
        f"Activation quant calibration type: {act_quant_calib_config} - "
        f"Weight quant calibration type: {args.weight_quant_calibration_type} - "
        f"Calibrate BN: {args.calibrate_bn} - "
        f"Channel Splitting Ratio: {args.channel_splitting_ratio} - "
        f"Split Input: {args.channel_splitting_split_input} - "
        f"Merge BN: {args.merge_bn}")

    # Get model-specific configurations about input shapes and normalization
    model_config = get_model_config(args.model_name)

    # Generate calibration and validation dataloaders
    resize_shape = model_config['resize_shape']
    center_crop_shape = model_config['center_crop_shape']
    inception_preprocessing = model_config['inception_preprocessing']
    calib_loader = generate_dataloader(
        args.calibration_dir,
        args.batch_size_calibration,
        args.workers,
        resize_shape,
        center_crop_shape,
        args.calibration_samples,
        inception_preprocessing)
    val_loader = generate_dataloader(
        args.validation_dir,
        args.batch_size_validation,
        args.workers,
        resize_shape,
        center_crop_shape,
        inception_preprocessing=inception_preprocessing)

    # Get the model from torchvision
    model = get_torchvision_model(args.model_name)
    model = model.to(dtype)
    model.eval()

    # Preprocess the model for quantization
    if args.target_backend == 'flexml':
        # flexml requires static shapes, pass a representative input in
        img_shape = model_config['center_crop_shape']
        model = preprocess_for_flexml_quantize(
            model,
            torch.ones(1, 3, img_shape, img_shape, dtype=dtype),
            equalize_iters=args.graph_eq_iterations,
            equalize_merge_bias=args.graph_eq_merge_bias,
            merge_bn=not args.calibrate_bn)
    elif args.target_backend == 'fx' or args.target_backend == 'layerwise':
        model = preprocess_for_quantize(
            model,
            equalize_iters=args.graph_eq_iterations,
            equalize_merge_bias=args.graph_eq_merge_bias,
            merge_bn=args.merge_bn,
            channel_splitting_ratio=args.channel_splitting_ratio,
            channel_splitting_split_input=args.channel_splitting_split_input)
    else:
        raise RuntimeError(f"{args.target_backend} backend not supported.")

    # If available, use the selected GPU
    if args.gpu is not None:
        torch.cuda.set_device(args.gpu)
        model = model.cuda(args.gpu)
        cudnn.benchmark = False

    if args.act_equalization is not None:
        print("Applying activation equalization:")
        apply_act_equalization(model, calib_loader, layerwise=args.act_equalization == 'layerwise')
    device = next(iter(model.parameters())).device
    # Define the quantized model
    quant_model = quantize_model(
        model,
        dtype=dtype,
        device=device,
        backend=args.target_backend,
        scale_factor_type=args.scale_factor_type,
        bias_bit_width=args.bias_bit_width,
        weight_bit_width=args.weight_bit_width,
        weight_narrow_range=args.weight_narrow_range,
        weight_param_method=args.weight_quant_calibration_type,
        weight_quant_granularity=args.weight_quant_granularity,
        act_quant_granularity=args.act_quant_granularity,
        weight_quant_type=args.weight_quant_type,
        layerwise_first_last_bit_width=args.layerwise_first_last_bit_width,
        act_bit_width=args.act_bit_width,
        act_param_method=args.act_quant_calibration_type,
        act_quant_percentile=args.act_quant_percentile,
        act_quant_type=args.act_quant_type,
        quant_format=args.quant_format,
        layerwise_first_last_mantissa_bit_width=args.layerwise_first_last_mantissa_bit_width,
        layerwise_first_last_exponent_bit_width=args.layerwise_first_last_exponent_bit_width,
        weight_mantissa_bit_width=args.weight_mantissa_bit_width,
        weight_exponent_bit_width=args.weight_exponent_bit_width,
        act_mantissa_bit_width=args.act_mantissa_bit_width,
        act_exponent_bit_width=args.act_exponent_bit_width,
        act_scale_computation_type=args.act_scale_computation_type,
        uint_sym_act_for_unsigned_values=args.uint_sym_act_for_unsigned_values)

    if args.act_scale_computation_type == 'static':
        # Calibrate the quant_model on the calibration dataloader
        print("Starting activation calibration:")
        calibrate(calib_loader, quant_model)

    if args.gpfq:
        print("Performing GPFQ:")
        apply_gpfq(
            calib_loader,
            quant_model,
            p=args.gpfq_p,
            act_order=args.gpxq_act_order,
            compression_rate=args.compression_rate)

    if args.gpfa2q:
        print("Performing GPFA2Q:")
        apply_gpfq(
            calib_loader,
            quant_model,
            p=args.gpfq_p,
            act_order=args.gpxq_act_order,
            use_gpfa2q=args.gpfa2q,
            accumulator_bit_width=args.accumulator_bit_width,
            compression_rate=args.compression_rate)

    if args.gptq:
        print("Performing GPTQ:")
        apply_gptq(calib_loader, quant_model, act_order=args.gpxq_act_order)

    if args.learned_round:
        print("Applying Learned Round:")
        apply_learned_round_learning(
            quant_model,
            calib_loader,
            iters=args.learned_round_iters,
            optimizer_lr=args.learned_round_lr)

    if args.calibrate_bn:
        print("Calibrate BN:")
        calibrate_bn(calib_loader, quant_model)

    if args.bias_corr:
        print("Applying bias correction:")
        apply_bias_correction(calib_loader, quant_model)

    # Validate the quant_model on the validation dataloader
    print("Starting validation:")
    with torch.no_grad(), quant_inference_mode(quant_model):
        param = next(iter(quant_model.parameters()))
        device, dtype = param.device, param.dtype
        ref_input = generate_ref_input(args, device, dtype)
        quant_model(ref_input)
        compiled_model = torch.compile(quant_model, fullgraph=True, disable=not args.compile)
        validate(val_loader, compiled_model, stable=dtype != torch.bfloat16)

    if args.export_onnx_qcdq or args.export_torch_qcdq:
        # Generate reference input tensor to drive the export process
        param = next(iter(quant_model.parameters()))
        device, dtype = param.device, param.dtype
        ref_input = generate_ref_input(args, device, dtype)

        export_name = os.path.join(args.export_dir, config)
        if args.export_onnx_qcdq:
            export_name = export_name + '.onnx'
            export_onnx_qcdq(
                quant_model, ref_input, export_name, opset_version=args.onnx_opset_version)
        if args.export_torch_qcdq:
            export_name = export_name + '.pt'
            export_torch_qcdq(quant_model, ref_input, export_name)


if __name__ == '__main__':
    main()



# Copyright (c) 2024, Advanced Micro Devices, Inc.
# Copyright (c) 2017, liukuang
# All rights reserved.
# SPDX-License-Identifier: MIT

import torch.nn as nn
import torch.nn.functional as F

from brevitas.nn.quant_layer import WeightQuantType
from brevitas.quant import Int8WeightPerChannelFloat
from brevitas_examples.bnn_pynq.models.resnet import QuantBasicBlock
from brevitas_examples.bnn_pynq.models.resnet import QuantResNet


def weight_init(layer):
    if isinstance(layer, nn.Conv2d):
        nn.init.kaiming_normal_(layer.weight, nn.init.calculate_gain('relu'))
        if layer.bias is not None:
            layer.bias.data.zero_()
    elif isinstance(layer, nn.BatchNorm2d):
        nn.init.constant_(layer.weight, 1)
        nn.init.constant_(layer.bias, 0)


class BasicBlock(nn.Module):
    """Basic block architecture modified for CIFAR10.
    Adapted from https://github.com/kuangliu/pytorch-cifar"""
    expansion = 1

    def __init__(self, in_planes: int, planes: int, stride: int = 1):
        super(BasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(
            in_planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)

        self.downsample = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            # using a convolution shortcut rather than identity
            self.downsample = nn.Sequential(
                nn.Conv2d(
                    in_planes,
                    self.expansion * planes,
                    kernel_size=1,
                    stride=stride,
                    bias=False,
                ),
                nn.BatchNorm2d(self.expansion * planes),
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.downsample(x)
        out = F.relu(out)
        return out


class ResNet(nn.Module):
    """ ResNet architecture modified for CIFAR10.
    Adapted from https://github.com/kuangliu/pytorch-cifar"""

    def __init__(self, block_impl, num_blocks, num_classes: int = 10):
        super(ResNet, self).__init__()

        # stride and padding of 1 with kernel size of 3, compared to ImageNet model
        self.conv1 = nn.Conv2d(3, 64, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(64)

        self.in_planes = 64
        self.layer1 = self.create_block(block_impl, 64, num_blocks[0], stride=1)
        self.layer2 = self.create_block(block_impl, 128, num_blocks[1], stride=2)
        self.layer3 = self.create_block(block_impl, 256, num_blocks[2], stride=2)
        self.layer4 = self.create_block(block_impl, 512, num_blocks[3], stride=2)
        self.linear = nn.Linear(512 * block_impl.expansion, num_classes)

        self.apply(weight_init)

    def create_block(self, block_impl, planes, num_blocks, stride):
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for stride in strides:
            layers.append(block_impl(self.in_planes, planes, stride))
            self.in_planes = planes * block_impl.expansion  # expand input planes
        return nn.Sequential(*layers)

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out = F.avg_pool2d(out, 4)
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        return out


def float_resnet18(num_classes: int = 10) -> ResNet:
    model = ResNet(BasicBlock, [2, 2, 2, 2], num_classes)
    return model


def quant_resnet18(
        num_classes: int = 10,
        act_bit_width: int = 8,
        acc_bit_width: int = 32,
        weight_bit_width: int = 8,
        weight_quant: WeightQuantType = Int8WeightPerChannelFloat) -> QuantResNet:
    weight_quant = weight_quant.let(accumulator_bit_width=acc_bit_width)
    model = QuantResNet(
        block_impl=QuantBasicBlock,
        num_blocks=[2, 2, 2, 2],
        num_classes=num_classes,
        act_bit_width=act_bit_width,
        weight_bit_width=weight_bit_width,
        weight_quant=weight_quant,
        last_layer_weight_quant=Int8WeightPerChannelFloat,
        first_maxpool=False,
        zero_init_residual=False,
        round_average_pool=False)
    return model



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC
from abc import abstractmethod
from copy import deepcopy
from dataclasses import dataclass
from dataclasses import field
from functools import partial
from operator import attrgetter
from typing import List, Optional, Set
import warnings

import torch
from torch.fx import GraphModule as TorchGraphModule

from brevitas.fx import GraphModule
from brevitas.graph.calibrate import disable_return_quant_tensor
from brevitas.graph.calibrate import DisableEnableQuantization
from brevitas.graph.calibrate import restore_return_quant_tensor
from brevitas.graph.utils import is_conv_transposed
import brevitas.nn as qnn
from brevitas.quant_tensor import IntQuantTensor
from brevitas.utils.quant_utils import _CachedIO

SUPPORTED_TCONV_OP = (qnn.QuantConvTranspose1d, qnn.QuantConvTranspose2d, qnn.QuantConvTranspose3d)

SUPPORTED_CONV_OP = (qnn.QuantConv1d, qnn.QuantConv2d, qnn.QuantConv3d, *SUPPORTED_TCONV_OP)


class StopFwdException(Exception):
    pass


@dataclass
class LayerHandler:
    layer_names: Set = field(default_factory=set)
    forward_count: int = 0


class gpxq_mode(ABC):
    """
    Apply GPxQ algorithm.

    Args:
        model (Module): The model to quantize with GPxQ
        group_of_parallel_layers (Optional, List[str]): .List of lists where each inner list is a group
            of layer names that can be optimized in parallel. Default: None
        inplace (bool): Wheter to apply GPFQ inplace or perform a deepcopy. Default: True
        create_weight_orig (bool): If True, store the original floating point weights before applying
            gpxq. These weights will be used anytime quantization is disabled. Default: True
        use_quant_activations (bool): Wheter to leave quantize activations enabled while performing
            GPxQ. Default: False
        act_order (bool): Whether to order greedy path following by Hessian approximation. Default: False
        return_forward_output (bool): If True, returns the output of the forward pass. Otherwise the
            forward call inside the context manager returns None. Default: False

    Example:
        >>> with torch.no_grad():
        >>>     with gpxq_mode(model) as gpxq:
        >>>         gpxq_mode = gpxq.model
        >>>         for i in tqdm(range(gpxq.num_layers)):
        >>>             for img, t in calib_loader:
        >>>                 img = img.cuda()
        >>>                 gpxq_mode(img)
        >>>             gpxq.update()
    """

    def __init__(
            self,
            model,
            group_of_parallel_layers: Optional[List[str]] = None,
            inplace: bool = True,
            create_weight_orig: bool = True,
            use_quant_activations: bool = True,
            act_order: bool = False,
            return_forward_output: bool = False) -> None:

        if not inplace:
            model = deepcopy(model)
        self.model = model
        self.create_weight_orig = create_weight_orig
        self.use_quant_activations = use_quant_activations
        self.hook_dict = dict()
        self.gpxq_layers = dict()
        # reference for each layer to update
        self.current_layer = LayerHandler()
        # How many layer to optimize
        self.num_layers = 0
        # Quantize following magnitude of activation
        self.act_order = act_order
        # How many subblock to use during GPTQ for each layer

        self.disable_quant_inference = DisableEnableQuantization()
        self.return_quant_tensor_state = dict()

        self.group_of_parallel_layers = group_of_parallel_layers
        self.return_forward_output = return_forward_output

        self.orig_forward = self.model.forward
        if isinstance(self.model, (GraphModule, TorchGraphModule)):
            self.model.__class__.forward = self.catch_stopfwd
        else:
            self.model.forward = self.catch_stopfwd

    def _is_module_supported(self, module):
        if isinstance(module, SUPPORTED_CONV_OP):
            return True
        elif isinstance(module, qnn.QuantLinear):
            return True
        else:
            return False

    def __enter__(self):
        # The user can specify on which layers to apply gptq in parallel.
        # All the others will be executed sequentially
        dict_of_layers = {
            name: [(name, module)] for name,
            module in self.model.named_modules() if self._is_module_supported(module)}
        if self.group_of_parallel_layers is not None:
            for parallel_layers in self.group_of_parallel_layers:
                for name in parallel_layers:
                    if name not in dict_of_layers:
                        raise ValueError(
                            "The layer {} is not present in the model or it is not supported for GPTQ"
                            .format(name))
                    del dict_of_layers[name]
                names = '_'.join(parallel_layers)
                dict_of_layers[names] = [
                    (name, attrgetter(name)(self.model)) for name in parallel_layers]

        # Print warning if hooks are attached to any module, since the normal forward flow of the
        # network is highly disrupted during GPxQ
        for _, parallel_layers in dict_of_layers.items():
            for name, module in parallel_layers:
                if len(module._forward_hooks) > 0 or len(module._forward_pre_hooks):
                    warnings.warn(
                        f'Hooks detected during setup for GPxQ. '
                        f'Behaviour might deviate from what expected.')

                # Attach hooks for GPTQ
                if self._is_module_supported(module):
                    gpxq_module_optimizer = self.initialize_module_optimizer(
                        module,
                        name,
                        act_order=self.act_order,
                        len_parallel_layers=len(parallel_layers),
                        create_weight_orig=self.create_weight_orig)
                    hook_fn = partial(
                        gpxq_module_optimizer.update_batch, current_layer=self.current_layer)
                    self.hook_dict[name] = module.register_forward_pre_hook(hook_fn)
                    self.gpxq_layers[name] = gpxq_module_optimizer

        if not self.use_quant_activations:
            self.return_quant_tensor_state = disable_return_quant_tensor(self.model)
            self.disable_quant_inference.disable_act_quantization(
                self.model, is_training=self.model.training)
            self.disable_quant_inference.disable_bias_quantization(
                self.model, is_training=self.model.training)

        self.num_layers = len(dict_of_layers)
        return self

    def __exit__(self, type, value, traceback):
        if isinstance(self.model, (GraphModule, TorchGraphModule)):
            self.model.__class__.forward = self.orig_forward
        else:
            self.model.forward = self.orig_forward

        if not self.use_quant_activations:
            self.disable_quant_inference.enable_act_quantization(
                self.model, is_training=self.model.training)
            self.disable_quant_inference.enable_bias_quantization(
                self.model, is_training=self.model.training)
            restore_return_quant_tensor(self.model, self.return_quant_tensor_state)

    def update(self):
        for name in self.current_layer.layer_names:
            self.gpxq_layers[name].single_layer_update()
            self.hook_dict[name].remove()
        self.current_layer.layer_names.clear()

    @abstractmethod
    def catch_stopfwd(self, *args, **kwargs):
        pass


class GPxQ(ABC):

    def __init__(
            self, layer, name, act_order, len_parallel_layers=1, create_weight_orig=True) -> None:
        self.layer = layer
        self.name = name
        self.act_order = act_order
        if self.layer.weight_quant.is_groupwise:
            weight = self.layer.weight_quant.apply_input_view(self.layer.weight)
            weight = weight.view(self.layer.weight_quant.quant_injector.reshaped_groupwise_shape)
            self.layer.weight.data = weight.data
            self.layer.in_channels = weight.shape[1] if is_conv_transposed(
                self.layer) else weight.shape[0]

        weight_shape = torch.tensor(layer.weight.shape)

        if create_weight_orig and not hasattr(self.layer, 'weight_orig'):
            self.layer.register_buffer('weight_orig', layer.weight.detach().clone())

        # By default, use groups = 1
        self.groups = 1
        if isinstance(self.layer, SUPPORTED_CONV_OP):
            if is_conv_transposed(self.layer):
                weight_shape[1], weight_shape[0] = weight_shape[0], weight_shape[1]
            self.groups = self.layer.groups

        # Number of rows is equal to the output channels (OC)
        self.rows = weight_shape[0]
        # Number of columns is equal to the input channels (IC)
        self.columns = torch.prod(weight_shape[1:])
        self.len_parallel_layers = len_parallel_layers

        self.disable_pre_forward_hook = False
        # Some layers require knowledge from quant inputs to compute quant weights
        self.quant_metadata = None

    def process_input(self, inp):
        # Input is a tuple, so we take first element
        inp = inp[0]
        inp = self.layer.input_quant(inp)

        is_quant_enabled = self.layer.weight_quant.is_quant_enabled

        # If using quantized activations, inp could be IntQuantTensor. In
        # this case, we overwrite the metadata.
        if isinstance(inp, IntQuantTensor):
            if is_quant_enabled and self.quant_metadata is None:
                self.quant_metadata = _CachedIO(inp, metadata_only=True)
            inp = inp.value

        # If input is unbatched, add batch_size = 1
        if len(inp.shape) == 1:
            warnings.warn("Found unbatched input, adding batch dimension equal to 1")
            inp = inp.unsqueeze(0)

        # Define batch size before re-organizing the input
        if hasattr(inp, 'names') and 'N' in inp.names:
            batch_dim = inp.names.index('N')
            inp.rename_(None)
            inp = inp.transpose(0, batch_dim)

        return inp

    @abstractmethod
    def update_batch(self):
        pass

    @abstractmethod
    def single_layer_update(self):
        pass

    def get_quant_weights(self, i, i1, permutation_list):
        # We need to recompute quant weights at runtime since our float weights are being updated
        # Add offset in case of blockwise computation
        i = i1 + i
        # For QuantLinear and for some QuantConvolutional layers, we exploit the possibility
        # of quantizing only a subset of the entire matrix speeding up the computation of GPxQ
        if isinstance(self.layer, qnn.QuantLinear):
            if self.layer.weight_quant.is_groupwise:
                # No slicing, not optimized
                index = permutation_list[0][i]
                q = self.layer.quant_weight(quant_input=self.quant_metadata).value.unsqueeze(
                    0)  # [1, OC, 1]
                q = q[:, :, index:index + 1]  # [groups, OC/groups, 1]
            else:
                index = permutation_list[0][i]
                subtensor_slice_list = [None, (index, index + 1)]
                q = self.layer.quant_weight(
                    subtensor_slice_list=subtensor_slice_list,
                    quant_input=self.quant_metadata).value.unsqueeze(0)  # [1, OC, 1]
        elif isinstance(self.layer, SUPPORTED_CONV_OP):
            # For depthwise and ConvTranspose we fall back to quantizing the entire martix.
            # For all other cases, we create a mask that represent the slicing we will perform on the weight matrix
            # and we quantize only the selected dimensions.
            if self.layer.weight_quant.is_groupwise or self.groups > 1 or (
                    self.groups == 1 and
                    isinstance(self.layer, (qnn.QuantConvTranspose1d, qnn.QuantConvTranspose2d))):

                quant_weight = self.layer.quant_weight(quant_input=self.quant_metadata)
                quant_weight = quant_weight.value

                if isinstance(self.layer, (qnn.QuantConvTranspose1d, qnn.QuantConvTranspose2d)):
                    quant_weight = quant_weight.transpose(1, 0)  # This performs a view
                quant_weight = quant_weight.flatten(1)
                quant_weight = quant_weight.view(self.groups, -1, quant_weight.shape[-1])

                if self.act_order:
                    for ii, perm in enumerate(permutation_list):
                        quant_weight[ii, :, :] = quant_weight[ii, :, perm]

                q = quant_weight[:, :, i:i + 1]  # [groups, OC/groups, 1]
            else:
                index = permutation_list[0][i]
                shapes = self.layer.weight.shape[1:]
                index_2d_to_nd = []
                residual_index = index.item()
                for shape in shapes[::-1]:
                    index_2d_to_nd.append((residual_index % shape, residual_index % shape + 1))
                    residual_index = residual_index // shape
                index_2d_to_nd = index_2d_to_nd[::-1]
                index_2d_to_nd.insert(0, None)
                q = self.layer.quant_weight(
                    subtensor_slice_list=index_2d_to_nd,
                    quant_input=self.quant_metadata).value.flatten(1)  # [OC, 1]
                q = q.unsqueeze(0)  # [1, OC, 1]
        # We need to remove the last dim
        q = q.squeeze(2)  # [groups, OC/groups] or [1, OC]
        return q



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABCMeta
from abc import abstractmethod
from typing import List, Optional, Tuple, Type, Union

from brevitas.inject import ExtendedInjector
from brevitas.inject import Injector
from brevitas.proxy.parameter_quant import BiasQuantProxyProtocol
from brevitas.proxy.parameter_quant import WeightQuantProxyProtocol
from brevitas.quant import NoneBiasQuant
from brevitas.quant import NoneWeightQuant
from brevitas.quant_tensor import QuantTensor

from .base import QuantProxyMixin

WeightQuantType = Union[WeightQuantProxyProtocol, Type[Injector], Type[ExtendedInjector]]
BiasQuantType = Union[BiasQuantProxyProtocol, Type[Injector], Type[ExtendedInjector]]


class QuantWeightMixin(QuantProxyMixin):
    __metaclass__ = ABCMeta

    def __init__(self, weight_quant: Optional[WeightQuantType], **kwargs):
        QuantProxyMixin.__init__(
            self,
            quant=weight_quant,
            proxy_protocol=WeightQuantProxyProtocol,
            none_quant_injector=NoneWeightQuant,
            kwargs_prefix='weight_',
            proxy_prefix='weight_',
            **kwargs)
        self._cached_sub_tensor_slice_list_modules = None

    @property
    @abstractmethod
    def output_channel_dim(self) -> int:
        pass

    def quant_weight(
            self,
            quant_input: Optional[QuantTensor] = None,
            subtensor_slice_list: List[Optional[Tuple[int, int]]] = None):
        weights_to_quantize = self.weight
        if not self.weight_quant.is_quant_enabled and hasattr(self, 'weight_orig'):
            weights_to_quantize = self.weight_orig
        if subtensor_slice_list is not None:
            # prepare the quantizer for a subtensor input, if any modifications are required
            # we set a list of tuples rather than a list of slices so that it's jit friendly
            # slices generation is handled by each module internally

            # we cache which modules require the attribute
            if self._cached_sub_tensor_slice_list_modules is not None:
                for m in self._cached_sub_tensor_slice_list_modules:
                    m.subtensor_slice_list = subtensor_slice_list
            else:
                self._cached_sub_tensor_slice_list_modules = []
                for m in self.weight_quant.modules():
                    if hasattr(m, 'subtensor_slice_list'):
                        self._cached_sub_tensor_slice_list_modules.append(m)
                        m.subtensor_slice_list = subtensor_slice_list
            # generate slices for the weight tensor based on the list passed in
            weight_slice_tuple = tuple(
                slice(*s) if s is not None else slice(s) for s in subtensor_slice_list)
        else:
            weight_slice_tuple = slice(None)
        if self.weight_quant.requires_quant_input:
            out = self.weight_quant(weights_to_quantize[weight_slice_tuple], quant_input)
        else:
            out = self.weight_quant(weights_to_quantize[weight_slice_tuple])
        if subtensor_slice_list is not None:
            # Restore the quantizer behaviour to full tensor quantization
            # The modules to slice should have been cached already at this point
            assert self._cached_sub_tensor_slice_list_modules is not None, "Missing cache of modules to slice."
            for m in self._cached_sub_tensor_slice_list_modules:
                m.subtensor_slice_list = [None]
        return out

    def register_parameter(self, name, value):
        super(QuantWeightMixin, self).register_parameter(name, value)
        if hasattr(self, 'weight_quant') and name == 'weight':
            self.weight_quant.init_tensor_quant()
            self.weight_quant.to(self.weight.device)


class QuantBiasMixin(QuantProxyMixin):
    __metaclass__ = ABCMeta

    def __init__(
            self,
            bias_quant: Optional[BiasQuantType],
            cache_inference_bias: bool = False,
            **kwargs):
        QuantProxyMixin.__init__(
            self,
            quant=bias_quant,
            proxy_protocol=BiasQuantProxyProtocol,
            none_quant_injector=NoneBiasQuant,
            kwargs_prefix='bias_',
            proxy_prefix='bias_',
            **kwargs)

    def quant_bias(self):
        if self.bias is None:
            return None
        quant_bias = self.bias_quant(self.bias)
        return quant_bias

    def register_parameter(self, name, value):
        super(QuantBiasMixin, self).register_parameter(name, value)
        if hasattr(self, 'bias_quant') and name == 'bias':
            self.bias_quant.init_tensor_quant()
            self.bias_quant.to(self.bias.device)



"""
Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

Based on https://github.com/nod-ai/SHARK/blob/main/apps/language_models/scripts/sharded_vicuna_fp32.py

Copyright 2023 Nod.ai

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.


---- LLVM Exceptions to the Apache 2.0 License ----

As an exception, if, as a result of your compiling your source code, portions
of this Software are embedded into an Object form of such source code, you
may redistribute such embedded portions in such Object form without complying
with the conditions of Sections 4(a), 4(b) and 4(d) of the License.

In addition, if you combine or link compiled forms of this Software with
software that is licensed under the GPLv2 ("Combined Software") and if a
court of competent jurisdiction determines that the patent provision (Section
3), the indemnity provision (Section 9) or other Section of the License
conflicts with the conditions of the GPLv2, you may retroactively and
prospectively choose to deem waived or otherwise exclude such Section(s) of
the License, but only in their entirety and only with respect to the Combined
Software.
"""
import argparse
from io import BytesIO
from pathlib import Path
import re
from typing import List

import torch
from torch._decomp import get_decompositions
import torch_mlir
from torch_mlir import TensorPlaceholder
from tqdm import tqdm

from brevitas.backport.fx._symbolic_trace import wrap
from brevitas.backport.fx.experimental.proxy_tensor import make_fx
from brevitas_examples.llm.llm_quant.export import block_quant_layer_level_manager
from brevitas_examples.llm.llm_quant.export import BlockQuantProxyLevelManager
from brevitas_examples.llm.llm_quant.export import brevitas_layer_export_mode
from brevitas_examples.llm.llm_quant.export import brevitas_proxy_export_mode
from brevitas_examples.llm.llm_quant.export import LinearWeightBlockQuantHandler
from brevitas_examples.llm.llm_quant.export import replace_call_fn_target
from brevitas_examples.llm.llm_quant.mlir_custom_mm import brevitas_matmul_rhs_group_quant_library


# Due a tracing issue this annotation needs to be
# in the same module (== file) from which make_fx is called
# We also can't directly annotate torch.ops.quant.matmul_rhs_group_quant
# and so we trace a placeholder first and then replace it post tracing
@wrap(visible_to_make_fx=True)
def matmul_rhs_group_quant_placeholder(*args, **kwargs):
    return torch.ops.quant.matmul_rhs_group_quant(*args, **kwargs)


class LinearWeightBlockQuantHandlerFwd(LinearWeightBlockQuantHandler):

    def forward(self, x):
        # Due a tracing issue the call to this fn needs to be
        # in the same module (== file) from which make_fx is called
        out = matmul_rhs_group_quant_placeholder(
            x, self.int_weight, self.scale, self.zero_point, self.bit_width, self.group_size)
        if self.bias is not None:
            out = out + self.bias.view(1, -1)
        return out


class FirstVicunaLayer(torch.nn.Module):

    def __init__(self, model):
        super().__init__()
        self.model = model

    def forward(self, hidden_states, attention_mask, position_ids):
        outputs = self.model(
            hidden_states, attention_mask=attention_mask, position_ids=position_ids, use_cache=True)
        next_hidden_states = outputs[0]
        past_key_value_out0, past_key_value_out1 = (outputs[-1][0], outputs[-1][1])

        return (next_hidden_states, past_key_value_out0, past_key_value_out1)


class SecondVicunaLayer(torch.nn.Module):

    def __init__(self, model):
        super().__init__()
        self.model = model

    def forward(
            self, hidden_states, attention_mask, position_ids, past_key_value0, past_key_value1):
        outputs = self.model(
            hidden_states,
            attention_mask=attention_mask,
            position_ids=position_ids,
            past_key_value=(past_key_value0, past_key_value1),
            use_cache=True)
        next_hidden_states = outputs[0]
        past_key_value_out0, past_key_value_out1 = (outputs[-1][0], outputs[-1][1])

        return (next_hidden_states, past_key_value_out0, past_key_value_out1)


def write_in_dynamic_inputs0(module, dynamic_input_size):
    new_lines = []
    for line in module.splitlines():
        line = re.sub(f"{dynamic_input_size}x", "?x", line)
        if "?x" in line:
            line = re.sub("tensor.empty\(\)", "tensor.empty(%dim)", line)
        line = re.sub(f" {dynamic_input_size},", " %dim,", line)
        if "tensor.empty" in line and "?x?" in line:
            line = re.sub("tensor.empty\(%dim\)", "tensor.empty(%dim, %dim)", line)
        if "arith.cmpi" in line:
            line = re.sub(f"c{dynamic_input_size}", "dim", line)
        new_lines.append(line)
    new_module = "\n".join(new_lines)
    return new_module


def write_in_dynamic_inputs1(module, dynamic_input_size):
    new_lines = []
    for line in module.splitlines():
        if "dim_42 =" in line:
            continue
        if f"%c{dynamic_input_size}_i64 =" in line:
            new_lines.append("%dim_42 = tensor.dim %arg1, %c3 : tensor<1x1x1x?xf32>")
            new_lines.append(f"%dim_42_i64 = arith.index_cast %dim_42 : index to i64")
            continue
        line = re.sub(f"{dynamic_input_size}x", "?x", line)
        if "?x" in line:
            line = re.sub("tensor.empty\(\)", "tensor.empty(%dim_42)", line)
        line = re.sub(f" {dynamic_input_size},", " %dim_42,", line)
        if "tensor.empty" in line and "?x?" in line:
            line = re.sub(
                "tensor.empty\(%dim_42\)",
                "tensor.empty(%dim_42, %dim_42)",
                line,
            )
        if "arith.cmpi" in line:
            line = re.sub(f"c{dynamic_input_size}", "dim_42", line)
        new_lines.append(line)
    new_module = "\n".join(new_lines)
    return new_module


def compile_vicuna_layer(
    export_context_manager,
    export_class,
    vicuna_layer,
    hidden_states,
    attention_mask,
    position_ids,
    past_key_value0=None,
    past_key_value1=None,
):
    hidden_states_placeholder = TensorPlaceholder.like(hidden_states, dynamic_axes=[1])
    attention_mask_placeholder = TensorPlaceholder.like(attention_mask, dynamic_axes=[2, 3])
    position_ids_placeholder = TensorPlaceholder.like(position_ids, dynamic_axes=[1])

    if past_key_value0 is None and past_key_value1 is None:
        with export_context_manager(vicuna_layer, export_class):
            fx_g = make_fx(
                vicuna_layer,
                decomposition_table=get_decompositions([
                    torch.ops.aten.embedding_dense_backward,
                    torch.ops.aten.native_layer_norm_backward,
                    torch.ops.aten.slice_backward,
                    torch.ops.aten.select_backward,
                    torch.ops.aten.norm.ScalarOpt_dim,
                    torch.ops.aten.native_group_norm,
                    torch.ops.aten.upsample_bilinear2d.vec,
                    torch.ops.aten.split.Tensor,
                    torch.ops.aten.split_with_sizes,]),
            )(hidden_states, attention_mask, position_ids)
    else:
        with export_context_manager(vicuna_layer, export_class):
            fx_g = make_fx(
                vicuna_layer,
                decomposition_table=get_decompositions([
                    torch.ops.aten.embedding_dense_backward,
                    torch.ops.aten.native_layer_norm_backward,
                    torch.ops.aten.slice_backward,
                    torch.ops.aten.select_backward,
                    torch.ops.aten.norm.ScalarOpt_dim,
                    torch.ops.aten.native_group_norm,
                    torch.ops.aten.upsample_bilinear2d.vec,
                    torch.ops.aten.split.Tensor,
                    torch.ops.aten.split_with_sizes,]),
            )(hidden_states, attention_mask, position_ids, past_key_value0, past_key_value1)

    def _remove_nones(fx_g: torch.fx.GraphModule) -> List[int]:
        removed_indexes = []
        for node in fx_g.graph.nodes:
            if node.op == "output":
                assert (len(node.args) == 1), "Output node must have a single argument"
                node_arg = node.args[0]
                if isinstance(node_arg, (list, tuple)):
                    node_arg = list(node_arg)
                    node_args_len = len(node_arg)
                    for i in range(node_args_len):
                        curr_index = node_args_len - (i + 1)
                        if node_arg[curr_index] is None:
                            removed_indexes.append(curr_index)
                            node_arg.pop(curr_index)
                    node.args = (tuple(node_arg),)
                    break

        if len(removed_indexes) > 0:
            fx_g.graph.lint()
            fx_g.graph.eliminate_dead_code()
            fx_g.recompile()
        removed_indexes.sort()
        return removed_indexes

    def _unwrap_single_tuple_return(fx_g: torch.fx.GraphModule) -> bool:
        """
        Replace tuple with tuple element in functions that return one-element tuples.
        Returns true if an unwrapping took place, and false otherwise.
        """
        unwrapped_tuple = False
        for node in fx_g.graph.nodes:
            if node.op == "output":
                assert (len(node.args) == 1), "Output node must have a single argument"
                node_arg = node.args[0]
                if isinstance(node_arg, tuple):
                    if len(node_arg) == 1:
                        node.args = (node_arg[0],)
                        unwrapped_tuple = True
                        break

        if unwrapped_tuple:
            fx_g.graph.lint()
            fx_g.recompile()
        return unwrapped_tuple

    def transform_fx(fx_g):
        for node in fx_g.graph.nodes:
            if node.op == "call_function":
                if node.target in [torch.ops.aten.empty]:
                    # aten.empty should be filled with zeros.
                    with fx_g.graph.inserting_after(node):
                        new_node = fx_g.graph.call_function(torch.ops.aten.zero_, args=(node,))
                        node.append(new_node)
                        node.replace_all_uses_with(new_node)
                        new_node.args = (node,)
        fx_g.graph.lint()

    transform_fx(fx_g)
    replace_call_fn_target(
        fx_g, src=matmul_rhs_group_quant_placeholder, target=torch.ops.quant.matmul_rhs_group_quant)

    fx_g.recompile()
    removed_none_indexes = _remove_nones(fx_g)
    was_unwrapped = _unwrap_single_tuple_return(fx_g)

    fx_g.graph.set_codegen(torch.fx.graph.CodeGen())
    fx_g.recompile()

    print("FX_G recompile")

    def strip_overloads(gm):
        """
        Modifies the target of graph nodes in :attr:`gm` to strip overloads.
        Args:
            gm(fx.GraphModule): The input Fx graph module to be modified
        """
        for node in gm.graph.nodes:
            if isinstance(node.target, torch._ops.OpOverload):
                node.target = node.target.overloadpacket
        gm.recompile()

    strip_overloads(fx_g)
    ts_g = torch.jit.script(fx_g)
    return ts_g


def compile_to_vmfb(inputs, layers, export_context_manager, export_class, is_first=True):
    mlirs = []
    for idx, layer in tqdm(enumerate(layers), desc="Getting mlirs"):
        if is_first:
            mlir_path = Path(f"{idx}_0.mlir")
            vmfb_path = Path(f"{idx}_0.vmfb")
        else:
            mlir_path = Path(f"{idx}_1.mlir")
            vmfb_path = Path(f"{idx}_1.vmfb")
        if vmfb_path.exists():
            continue
        if mlir_path.exists():
            f_ = open(mlir_path, "rb")
            bytecode = f_.read()
            f_.close()
        else:
            hidden_states_placeholder = TensorPlaceholder.like(inputs[0], dynamic_axes=[1])
            attention_mask_placeholder = TensorPlaceholder.like(inputs[1], dynamic_axes=[3])
            position_ids_placeholder = TensorPlaceholder.like(inputs[2], dynamic_axes=[1])
            if not is_first:
                pkv0_placeholder = TensorPlaceholder.like(inputs[3], dynamic_axes=[2])
                pkv1_placeholder = TensorPlaceholder.like(inputs[4], dynamic_axes=[2])
            print(f"Compiling layer {idx} mlir")
            if is_first:
                ts_g = compile_vicuna_layer(
                    export_context_manager, export_class, layer, inputs[0], inputs[1], inputs[2])
                module = torch_mlir.compile(
                    ts_g, (hidden_states_placeholder, inputs[1], inputs[2]),
                    output_type="torch",
                    backend_legal_ops=["quant.matmul_rhs_group_quant"],
                    extra_library=brevitas_matmul_rhs_group_quant_library,
                    use_tracing=False,
                    verbose=False)
            else:
                ts_g = compile_vicuna_layer(
                    export_context_manager,
                    export_class,
                    layer,
                    inputs[0],
                    inputs[1],
                    inputs[2],
                    inputs[3],
                    inputs[4])
                module = torch_mlir.compile(
                    ts_g,
                    (
                        inputs[0],
                        attention_mask_placeholder,
                        inputs[2],
                        pkv0_placeholder,
                        pkv1_placeholder),
                    output_type="torch",
                    backend_legal_ops=["quant.matmul_rhs_group_quant"],
                    extra_library=brevitas_matmul_rhs_group_quant_library,
                    use_tracing=False,
                    verbose=False)

            if is_first:
                module = write_in_dynamic_inputs0(str(module), 137)
                bytecode = module.encode("UTF-8")
                bytecode_stream = BytesIO(bytecode)
                bytecode = bytecode_stream.read()

            else:
                module = write_in_dynamic_inputs1(str(module), 138)
                if idx in [0, 5, 6, 7]:
                    module_str = module
                    module_str = module_str.splitlines()
                    new_lines = []
                    for line in module_str:
                        if len(line) < 1000:
                            new_lines.append(line)
                        else:
                            new_lines.append(line[:999])
                    module_str = "\n".join(new_lines)
                    f1_ = open(f"{idx}_1_test.mlir", "w+")
                    f1_.write(module_str)
                    f1_.close()

                bytecode = module.encode("UTF-8")
                bytecode_stream = BytesIO(bytecode)
                bytecode = bytecode_stream.read()

            f_ = open(mlir_path, "wb")
            f_.write(bytecode)
            f_.close()
        mlirs.append(bytecode)

    return mlirs


def sharded_weight_group_export(model, no_custom_packed_export):

    # SAMPLE_INPUT_LEN is used for creating mlir with dynamic inputs,
    # which is currently an increadibly hacky proccess
    # please don't change it
    SAMPLE_INPUT_LEN = 137

    placeholder_input0 = (
        torch.zeros([1, SAMPLE_INPUT_LEN, 4096]),
        torch.zeros([1, 1, SAMPLE_INPUT_LEN, SAMPLE_INPUT_LEN]),
        torch.zeros([1, SAMPLE_INPUT_LEN], dtype=torch.int64))

    placeholder_input1 = (
        torch.zeros([1, 1, 4096]),
        torch.zeros([1, 1, 1, SAMPLE_INPUT_LEN + 1]),
        torch.zeros([1, 1], dtype=torch.int64),
        torch.zeros([1, 32, SAMPLE_INPUT_LEN, 128]),
        torch.zeros([1, 32, SAMPLE_INPUT_LEN, 128]))

    if no_custom_packed_export:
        export_context_manager = brevitas_proxy_export_mode
        export_class = BlockQuantProxyLevelManager
    else:
        export_context_manager = brevitas_layer_export_mode
        # generate an export_class with the handler declared above
        export_class = block_quant_layer_level_manager(
            export_handlers=[LinearWeightBlockQuantHandlerFwd])

    layers0 = [FirstVicunaLayer(layer) for layer in model.model.layers]
    mlirs0 = compile_to_vmfb(
        placeholder_input0, layers0, export_context_manager, export_class, is_first=True)

    layers1 = [SecondVicunaLayer(layer) for layer in model.model.layers]
    mlirs1 = compile_to_vmfb(
        placeholder_input1, layers1, export_context_manager, export_class, is_first=False)
